/*

DESCRIPTION:    definitions for mpptproto display status with rs232-44780.
		This include file is shared between mpptproto and 
		rs232-44780.

COPYRIGHT:      (c) 2020, Brian Cornell, all rights reserved.

 TARGET MCU:    PIC16F18324, PIC16F1769

HW REVISION:    1

HW RELEASE:     prototype

VERSION:        0.1

MODIFICATION HISTORY:

    15 Feb 20, BWC, changed OUTL status code to OVCC.

    15 Oct 19, BWC, created file.


*/


#ifndef uint8_t
#include <stdint.h>
#endif


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * CONSTANTS NAMING CONVENTION.
 * 
 * S_<TYPE>_<INFO>, where TYPE=
 *  + DC=Duty Cycle
 *  + DF=Default value for a configurable parameter
 *  + EN=ENable bit
 *  + RT=RaTio or scaling factor
 *  + LM=LiMiting operational value
 *  + TM=TiMing
 *  + MP=Peripheral port bit MaPping
 *  + CH=character
 *  + IN=INitializing value
 *  + VL=VoLtage
 *  + QY=quantity (count) of something
 *  + CD=command data
 *  + BL=boolean value
 *  + MK=bit MasK
 *  + SZ=size of an object or data type
 * 
 * INFO is user defined and should be descriptive.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_CH_STX        0x02            //Start of text (prefix)
#define S_CH_ETX        0x03            //End of text (postfix)
#define S_CH_TERM       0xFF            //Line termination byte for rs232-44780
#define S_CD_GETD       "GZ"            //Get data payload command for mpptproto


#ifndef _EXCLUDE_OBJ

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * STATUS CODES AND STATE MANAGEMENT.
 * 
 * The constants STS_XXXX represent the status codes used throughout the 
 * control program and stored in object Config.sts as well as recorded in NVM.
 * The object StsList also provides brief text for status display via the 
 * service port.
 * 
 * STATE MANAGEMENT refers to identifying the operational state of the unit.
 *  + RUN: a non-error state that includes the status codes BOOT, STOP, and RUN.
 *  + IDLE: a non-operational state due to expected, recurring, environmental 
 *    conditions that prevent a RUN state.  These include low battery or PV 
 *    voltage.
 *  + FAULT: a non-recoverable error condition that requires intervention.  This
 *    state can only be exited by power cycle or reboot.
 * 
 * The constants RUN_STATE_IDLE and RUN_STATE_FAULT serve as limit tests that 
 * avoid the use of specific status codes to identify the operating state.
 * Codes less than RUN_STATE_IDLE are RUN states.  Codes less than 
 * RUN_STATE_FAULT are IDLE states and not recorded in NVM; status values 
 * greater than or equal to are true FAULT states and are recorded in NVM.
 *
 * These codes are mainly used by mpptproto, but rs232-44780 uses a few codes
 * for its own status display and to interpret status from mpptproto to control
 * the LCD backlight.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define STS_BOOT        0               //Booting
#define STS_STOP        1               //Stopped
#define STS_RUN         2               //Running
#define STS_BATL        3               //Low battery (also Vcc)
#define STS_PVVL        4               //Low PV voltage
#define STS_MCU         5               //MCU fault (stack, watchdog, etc.)
#define STS_NVM         6               //Invalid NVM
#define STS_OVCC        7               //Excessive over-current
#define STS_UND         8               //Undefined: must be last in list

#define RUN_STATE_IDLE  STS_BATL        //Codes < are RUN states
#define RUN_STATE_FAULT STS_MCU         //Codes < are IDLE states

const char *StsList[] = {
    {"BOOT"},
    {"STOP"},
    {"RUN"},
    {"BATL"},
    {"PVVL"},
    {"MCU"},
    {"NVM"},
    {"OVCC"},
    {"UNDF"}
};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * DISPLAY DATA.  The binary object dsp is sent from mpptproto to rs232-44780.
 *
 * Unlike the rest of mpptproto's console command set, which send ASCII 
 * responses, a binary data packet is sent for efficiency.  The object's last
 * member, term, contains the line termination byte that the RS232 driver 
 * interprets as 'end of line'.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
typedef struct {
    uint8_t     prefix;                 //Prefix marks beginning of data
    uint8_t     sts;                    //System status code; see status codes
    uint8_t     pwr;                    //Calculated PV power
    uint8_t     bat_v;                  //Battery volts
    uint8_t     bat_mv;                 //Battery mV
    uint8_t     pv_v;                   //PV volts
    uint8_t     pv_mv;                  //PV mV
    uint8_t     pv_a;                   //PV amps
    uint8_t     pv_ma;                  //PV mA
    uint8_t     postfix;                //Postfix marks end of data
    uint8_t     term;                   //Data termination byte
} dsp;

#endif



//End of file
