/*

DESCRIPTION:    Control firmware for Maximum Power Point Tracking (MPPT) 12V 
                Photo-Voltaic (PV) charger prototype.

                The purpose of this design is to develop, study, and refine the 
                Perturb & Observe (P&O) algorithm for use in a charger that 
                behaves like an automotive charging system: one that works to 
                maintain a constant bus voltage with the battery charge 
                dependent on load.

                This is done to support applications where the PV charging 
                system & battery are connected to constantly varying loads.
                Depending on the load and available PV energy, the charger may 
                support the entire load & charge the battery or support only the
                load (or partial) with no battery charge.  Hence, no battery 
                charging algorithm is included (nor intended) in the design.

                Specifications:
                    + PV input voltage: 15 - 25 volts
                    + PV input current: maximum 7A
                    + Charging voltage: 14V (adjustable)
                    + Charging current: 7.2A (max)
                    + Maximum continuous power: 100W
                    + Bias supply: 10 - 15V (provided by battery)
                    + MPPT algorithm: Perturb & Observe (P&O)
                    + Cooling: natural convection
                    + Operating temperature: -20 - +50C (not validated)

                This design is based on the 12 to 5V USB charger (125vbuck) and
                uses the same PCB, PIC, and switching core.  PCB modifications
                have been made with hand wire traces, trace cuts, and glued 
                components.

                The switching core of the charger remains a non-isolated 
                synchronous buck with limited low-side control.  Unlike 
                125vbuck, Current mode control with slope compensation is 
                employed.  A type 2 error amp in the outer control loop 
                regulates the charger voltage.

                An RS232 interface shared with the ICSP circuit provides a 
                console interface with a robust command set to allow 
                experimentation without firmware modification.  Many parameters
                are stored in Non-Volatile Memory (NVM).

                RS232 settings: 115.2kbps, 8 data bits, 1 stop, no parity

                This console port is accessed over the micro-USB connector on 
                the PCB and is optimized for use with a GBIB-style control 
                program.  A query command is also provided to support the 
                rs232-44780 display unit.  This command responds with a packet 
                of binary data that is rendered on a Hitachi 44780 style LCD 
                indicating unit status, battery, and PV output.

                The console / display interface is optional and the unit will
                operate without it provided that the proper options are set 
                (e.g. closed loop and autostart).

                A status LED on the PCB indicates if the charger is running 
                (Green) or in a stopped or fault state (RED).  There are no
                flash sequences used to communicate a status code.  The LED is
                managed by the Initialization and Charger on/off functions.

SYNTAX:         n/a

KNOWN BUGS:     n/a

COPYRIGHT:      (c) 2020, Brian Cornell, all rights reserved.

DEPENDENCIES:   see include file listing.

 TARGET MCU:    PIC16F1769

HW REVISION:    190613

HW RELEASE:     prototype

VERSION:        1.1

MODIFICATION HISTORY:

 
    15 Feb 20, BWC, removed low output voltage error handling and replaced with
    over-current check.  See in-line comments.

    6 Sep 19, BWC, created release.
    

*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * COMPILER & CLIB INCLUDES.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FIRMWARE INCLUDES:
 * + _xtal_freq.h:  sets the constant _XTAL_FREQ to the system oscillator
 *                  frequency (Fosc).  Used by the __delayxxx macros.
 * + binary.h:      Boolean constants shared across firmware projects.
 * + build.h:       managed by the script build.sh and called by the make file.
 *                  Maintains the compile build number.
 * + config.h:      PIC config registers
 * + dspstatus.h:   system status & display definitions.  Shared with project
 *                  rs232-44780 (used as optional status display).
 * + mpptproto.h:   main include file for this project.
 * + usart.h:       usart module function declarations, constants, etc.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <binary.h>
#include "config.h"
#ifndef _XTAL_FREQ
#include "_xtal_freq.h"
#endif
#include "usart.h"
#include "dspstatus.h"
#include "mpptproto.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * MODULE DATA.
 * + Config is a module-level object that maintains dynamic operating data (e.g.
 *   voltages, currents, etc.).
 * + noSample is used by the data collection block to indicate if valid data 
 *   exists in Config.pvv & Config.pvc.
 * + mpptIncrease is used by the MPP algorithm to indicate if the duty cycle 
 *   (DC) should increase (TRUE) or decrease (FALSE).
 * + mpptMove controls execution flow in the MPP algorithm: when TRUE the duty
 *   cycle is adjusted in the direction indicated by mpptIncrease.
 * + flag_ss is TRUE if a snapshot of system data exists in object snapshot.
 * + cal_req is TRUE if the system timer cal_secs has expired and indicates that
 *   PV current sense calibration is required.
 * + RestartTimer is maintained by the system timer and sets the delay before
 *   the autostart code block can execute when BOOT status is set.
 * + RDword enables word access to the ramData object for 16-bit NVM values.
 * + cal_secs is maintained by the system timer and sets the delay between 
 *   between PV current calibration events.
 * + count_mppt counts the number of iterations thru the RUN loop before the
 *   the MPP algorithm samples the PV power after adjusting the duty cycle.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
syscf Config;
bit noSample;
bit mpptIncrease;
bit mpptMove;
bit flag_ss;
bit cal_req;
uint8_t RestartTimer = 0;
uint16_t *RDword,
          cal_secs = 0,
          count_mppt = S_IN_MPPT;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Non-volatile parameter storage & management.  nvData resides in program flash
 * at the top of the high-endurance section (to make the lower portion 
 * contiguous & available for code).  ramData is the working copy in RAM.
 * 
 * Initialization of nvData & ramData MUST match: these are system defaults and 
 * enable recovery after NVM corruption or to restore via console commands.  
 * Refer to detailed explanation of NVM management at beginning of function 
 * Initialize.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const unsigned char nvData[FLASHROWSZ] @ 0x1FE0 = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};
unsigned char ramData[FLASHROWSZ] = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};



/*Function main.  Control program entry point (from root code) for Maximum Power 
 * Point Tracking (MPPT) 12V Photo-Voltaic (PV) charger prototype.  Manages 
 * boot, operation, fault handling, user interface, and shutdown.
 * 
 * Consists of two functional blocks:
 *  + Initialization
 *  + RUN loop
 * 
 * Initialization does just that:  initializes the MCU & operating environment 
 * and performs the necessary boot & diagnostic checks.  When done it always
 * turns control over to the RUN loop.
 * 
 * The RUN loop is an infinite loop that uses waterfall processing.  Several
 * discrete code blocks are ordered according to dependency and / or priority:
 *  + PV current sense calibration
 *  + Data collection
 *  + Fault monitoring
 *  + Safety shutdown
 *  + RUN monitoring & control
 *  + BOOT (autostart)
 *  + Console service
 * 
 * Each block is documented.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * FUNCTION DATA.
     * + cmdb points to the start of the console command buffer.
     * + cmdp points to the start of parameters (after the command) in the
     *   command buffer.
     * + cmd_ack points to the acknowledgement string to be sent in response to
     *   a command - either 'Ack' or 'Nck'.
     * + samplec counts the number of samples accumulated in the data collection
     *   block.
     * + counter is a general purpose loop counter and scratch variable.  It 
     *   WILL NOT be used for multiple iterations thru the RUN loop.
     * + count_batl counts the number of battery voltage samples that are below
     *   the minimum allowed value.  It is used to apply hysteresis in the
     *   sampling algorithm that controls low battery voltage shutdown.  Refer
     *   to the 'FAULT MONITORING & RECOVERY CONTROL' block for more detail.
     * + count_pvvl counts the number of PV voltage samples that are below
     *   the minimum allowed value.  It is used to apply hysteresis in the
     *   sampling algorithm that controls low PV voltage shutdown.  Refer
     *   to the FAULT 'MONITORING & RECOVERY CONTROL' block for more detail.
     * + count_ovcc counts the number of switching over-current events.  It is 
     *   used to apply hysteresis in the sampling algorithm that controls over-
     *   current shutdown.  Refer to the FAULT 'MONITORING & RECOVERY CONTROL' 
     *   block for more detail.
     * + reg_status is assigned the value of the STATUS register at boot.  Refer
     *   to the 'MCU FAILURE RECORDING' block for more detail.
     * + reg_pcon is assigned the value of the PCON register at boot.  Refer
     *   to the 'MCU FAILURE RECORDING' block for more detail.
     * + lock_mppt limits the amount of time (sleep iterations) that the MPP can
     *   remain at a given DC regardless measured power conditions.  Only used
     *   by the sweep algorithm.
     * + cmdi is the numeric value of the string pointed to by object cmdp.
     *   Used by the CONSOLE SERVICE block.
     * + pvvs_sum is the sum of PV voltage samples accumulated by the DATA
     *   COLLECTION block.
     * + pvcs_sum is the sum of PV current samples accumulated by the DATA
     *   COLLECTION block.
     * + pvvs is used by function ReadVoltage to store the per-sample PV 
     *   voltage.
     * + pvcs is used by function ReadVoltage to store the per-sample PV 
     *   current.
     * + snapshot is used by the SHUT-DOWN, BOOT, and CONSOLE SERVICE
     *   blocks to record dynamic voltage & current data based on an event or
     *   console command.
     * + ss_ptr points to either the Console or snapshot object.  Used by the
     *   CONSOLE SERVICE commands.
     * + lcd is a data object used by the CONSOLE SERVICE command defined in
     *   constant S_CD_GETD to send binary data to an external LCD display
     *   (RS232-44780).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    char *cmdb,                         //Start of command buffer
         *cmdp;                         //Start of command buffer parameters
    const char *cmd_ack;                //Pointer to command acknowledge string
    uint8_t samplec = S_IN_SAMPLES,     //Counts number of samples taken
            counter = 0,                //General use counter
            count_batl = S_LM_BATL_HYS, //BATL hysteresis counter
            count_pvvl = S_LM_PVVL_HYS, //PVVL hysteresis counter
            count_ovcc = 0,             //OVCC hysteresis counter
            reg_status,                 //Copy of STATUS register
            reg_pcon,                   //Copy of PCON register
            lock_mppt;
    unsigned int cmdi,                  //Console parameter as unsigned int
                 pvvs_sum = 0,          //Sum of adc values for PV volts
                 pvcs_sum = 0;          //Sum of adc values for PV current
    volts   pvvs,                       //Per sample PV volts
            pvcs;                       //Per sample PV current
    syscf   snapshot,                   //Snapshot of dynamic parameters
            *ss_ptr;                    //Points to snapshot or Config object
    dsp     lcd;                        //Data to send to LCD display


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * MCU FAILURE RECORDING.
     * 
     * If an MCU failure (e.g. stack overflow, watchdog, etc.) is detected at 
     * boot it's important to understand what caused the fault.  The STATUS &
     * PCON registers are saved at the start of the main function.  Once 
     * initialization is complete, but before the run loop, these are saved to 
     * object ramData for writing to NVM if the MCU status code is set.
     * 
     * They will remain in NVM until another MCU trap occurs, the NVM status
     * code is cleared, or defaults restored.  The console command for get NVM 
     * status will return the status code mnemonic along with the STATUS & PCON
     * registers, respectively, in hexadecimal.  Register values are only
     * recorded when an MCU status is set so these will be 0x00 for any other
     * code.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    reg_status = STATUS;
    reg_pcon = PCON;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZATION.
     * 
     * Function Initialize configures the MCU and peripherals.  Afterwards a 
     * delay provides settle time.  Local variables are initialized before
     * entering the RUN loop.
     * 
     * Initialize local variables:
     * + noSample is set to indicate that no active samples are populated in 
     *   their respective objects.  Until it is set, the run-time loop will not 
     *   execute beyond the data collection blocks.  This is by design and 
     *   ensures that the remainder of the loop is working with valid data.  
     *   Once the first valid samples are ready noSample is cleared and loop 
     *   execution proceeds normally.  This is a one-time event at boot.
     * + The mvr elements of objects pvvs & pvcs are set to high range for ADC
     *   acquisition.  These are used to obtain voltage & current averages of 
     *   the PV panel.
     * + Clear snapshot object.
     * + Assign pre/postfix & term constants to lcd object.  These are used for
     *   data validation and transmission termination, respectively, and don't 
     *   change.
     * + Save STATUS & PCON registers in ramData object for subsequent write to
     *   NVM if MCU error by the 'fault monitoring & recovery control' block
     *   within the run loop.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    __delay_ms(S_TM_INT);
    noSample = TRUE;
    pvvs.mvr = S_RT_ADC_MV_H;
    pvcs.mvr = pvvs.mvr;
    memset((void *) &snapshot, (int) 0x00, sizeof(syscf));
    lcd.prefix = S_CH_STX;
    lcd.postfix = S_CH_ETX;
    lcd.term = S_CH_TERM;

    if (Config.sts == STS_MCU)
    {
        ramData[FLASH_AR_STATUS] = (unsigned char) reg_status;
        ramData[FLASH_AR_PCON] = (unsigned char) reg_pcon;
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * RUN LOOP PRIORITIZATION & WATCHDOG MANAGEMENT.
     * 
     * Once entered, the run-time loop's only exit points are a reboot or power
     * cycle.  The loop consists of several discrete blocks responsible for 
     * data collection, monitoring, and control.  They are organized in order
     * of criticality and/or dependency of subsequent blocks.  Categorically the
     * order is:
     *      + PV current sense calibration
     *      + Data collection
     *      + Fault monitoring & recovery control (includes NVM status 
     *        recording)
     *      + Shut-down
     *      + Run monitoring & control (including MPPT algorithm)
     *      + Boot (autostart)
     *      + Console service
     * 
     * The watchdog timer is expected to be on in the run loop regardless of
     * state.  It should be set to a value that accounts for the longest loop
     * time with at least a 100% margin.  If it must be disabled in anticipation
     * of a long delay then it is the responsibility of that code block to 
     * re-enable.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SWDTEN = TRUE;

    while (TRUE)
    {
        CLRWDT();                           //Clear watchdog timer

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * PV CURRENT SENSE CALIBRATION.
         * 
         * The hall effect current sensor used on the PV input has a zero 
         * current offset of 2.5V that must be subtracted from the ADC value to
         * obtain an accurate current value.  The offset can also drift a small
         * amount over time, temperature, etc.  Hence, the calibration is done
         * at boot and then periodically.
         * 
         * The flag cal_req is set TRUE at boot and when a calibration is 
         * required.  The timer6 interrupt service routine maintains the seconds
         * counter cal_secs which approximates the number of elapsed seconds 
         * since boot or last calibration.  When cal_secs equals S_TM_CAL_PVC, 
         * flag cal_req is set TRUE.  It remains set until cleared by this code 
         * block.
         * 
         * For calibration the charger must be off so there's no current draw.
         * cal_req must be TRUE and the COG off for calibration to occur.  This 
         * criteria makes execution of this block variable in timing and 
         * dependent on a dormant period (e.g. nighttime).  It also means that
         * the value of S_TM_CAL_PVC isn't critical but should be set to a
         * value representing between 12~18 hours.  Low values might cause
         * multiple calibrations which doesn't hurt but isn't necessary.
         * 
         * A number of samples are taken to average out the noise and the result
         * is stored in the Config object.  The timer, cal_secs, is cleared 
         * along with the flag and must be done with interrupts off.
         * 
         * PV volt & current sample control is cleared and forced to resample 
         * with the new offset.  No alteration is made to the system status.
         * 
         * Since calibration is done with the charger off there should be no
         * safety or timing concerns.  The watchdog timer must be set to a value
         * high enough to provide a margin allowing for execution of this block
         * and the first pass thru the data collection block (which will loop &
         * reset WDT upon completion).
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((cal_req) && (COG1ASD0bits.ASE))
        {
            for (counter = 0, Config.pvco = 0; (counter < S_LM_CAL_PVC); \
                    ++counter)
            {
                ReadVoltage(PIN_IA_PVAMP, S_RT_PVC, &pvcs);
                Config.pvco += pvcs.adc;
            }
            Config.pvco /= S_LM_CAL_PVC;        //Store new offset

            di();
            cal_secs = 0;
            cal_req = FALSE;
            ei();

            //Reset sample control
            samplec = S_IN_SAMPLES;
            pvvs_sum = 0;
            pvcs_sum = pvvs_sum;
            noSample = TRUE;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * DATA COLLECTION.
         * 
         * This block measures system voltages and currents.  The PV current &
         * voltage are averaged using several samples because of their variable
         * impedance.  When the RUN loop is first entered at BOOT, flag nosample
         * is TRUE to ensure valid samples exist for all values before the
         * BOOT process can complete.  Steps:
         *  + Various system voltages are read first.  Those that require 
         *    averaging are stored in temporary objects for summing & subsequent 
         *    averaging.
         *  + Calculate average of measured samples for PV volts & current.  On 
         *    each loop sum the measured value and when sample count is reached
         *    average.
         *  + Calculate amperes: use CalcAvg to scale the voltage analog of PV 
         *    current to amps stored in standard volts object format.  
         *  + Calculate power: remove decimal portion and store integer value.
         *  + Reset counters for next sample set.  If this is the first sample
         *    set then clearing flag allows runtime loop to execute normally.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReadVoltage(PIN_IA_PVOLT, S_RT_PVV, &pvvs);
        ReadVoltage(PIN_IA_PVAMP, S_RT_PVC, &pvcs);
        ReadVoltage(PIN_IA_BVOLT, S_RT_BAT, &Config.bat);
        ReadVoltage(PIN_IA_OVOLT, S_RT_OUT, &Config.out);

        pvvs_sum += pvvs.adc;
        if (pvcs.adc < Config.pvco) pvcs.adc = Config.pvco;     //Noise compensation
        pvcs_sum += (pvcs.adc - Config.pvco);
        if (++samplec > S_LM_SAMPLES)
        {
            CalcAvg(pvvs_sum, S_LM_SAMPLES, S_RT_PVV, &Config.pvv);
            CalcAvg(pvcs_sum, S_LM_SAMPLES, S_RT_PVC, &Config.pvc);

            CalcAvg(Config.pvc.adc, S_QY_SAM_PVC, S_RT_MVA_PVC, &Config.pvc);
            Config.pwr = (uint8_t) ((Config.pvc.vmv * Config.pvv.vmv) / \
                    S_RT_VMV_MUL);

            samplec = S_IN_SAMPLES;
            pvvs_sum = 0;
            pvcs_sum = pvvs_sum;
            noSample = FALSE;
        }
        if (noSample) continue;         //Need valid sample @ boot to continue


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * FAULT MONITORING & RECOVERY CONTROL.
         * 
         * This block evaluates system values from the data collection block to
         * ensure that they are within limits for the current operating state,
         * and if not the unit's operating state is modified.  The current
         * operating state is determined using the STATE MANAGEMENT strategy
         * described in dspstatus.h:
         * 
         *  + RUN state values exceeding limits change the operating state to
         *    IDLE.
         *  + IDLE state values within limits change the operating state back to
         *    RUN.
         *  + FAULT state is established during the initialization, BOOT, or RUN
         *    blocks.  It's purpose here is to capture the status code in
         *    NVM (details below).
         * 
         * The RUN and IDLE state checks employ a simple hysteresis or PI 
         * algorithm that smooths operation and prevents state changes for 
         * instantaneous deviations that are typical in this type of system.
         * The amount of hysteresis can be adjusted by changing the values of 
         * the S_LM_xxxx_HYS constants.
         * 
         * Over-current events are monitored by checking for the number of
         * events in a single loop.  If this continues in successive loops a
         * FAULT code is set.  This is done for safety since it indicates 
         * either a HW failure or shorted output.  Additional explanation
         * can be found in the BOOT block of this function.
         * 
         * The minimum operating and restart values are defined with a set of
         * constants ending in _MIN & _RSRT respectively.  There must be 
         * sufficient separation between these to prevent RUN-IDLE oscillations.
         * 
         * NVM STATUS RECORDING.  Only FAULT status codes are recorded to NVM
         * since these are considered failures and disable the unit.  Only the 
         * first event of a specific status code will be recorded until a 
         * different status code presents.  Example:  the first MCU status
         * event will be recorded.  Subsequent MCU codes will not since the 
         * status is the same.  Before committing ensure that Vdd is greater 
         * than minimum to prevent NVM corruption if powering down.
         * 
         * A while loop is used to manage logic flow thru this block because it
         * eliminates nesting of if/else blocks.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifndef _LIMITS_OFF
        while (TRUE)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Monitor system values during RUN states.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (Config.sts < RUN_STATE_IDLE)
            {
                //Low battery voltage
                if (Config.bat.volt < S_VL_BAT_MIN)
                {
                    if ((++count_batl) > S_LM_BATL_HYS)
                    {
                        Config.sts = STS_BATL;
                        break;
                    }
                }
                else if (count_batl) --count_batl;

                //Low PV voltage
                if ((Config.pvv.volt < ramData[FLASH_AR_PVVL_STOP]) \
                        && (!(Config.pvc.vmv))) 
                {
                    if ((++count_pvvl) > S_LM_PVVL_HYS)
                    {
                        Config.sts = STS_PVVL;
                        break;
                    }
                }
                else if (count_pvvl) --count_pvvl;

                //Over-current
                if (Config.ovcc > S_LM_OVCC_CNT)
                {
                    Config.ovcc = 0;                    //Clear counter
                    if ((++count_ovcc) > S_LM_OVCC_HYS)
                    {
                        Config.sts = STS_OVCC;
                        break;
                    }
                }
                else if (count_ovcc) --count_ovcc;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Monitor system values during IDLE state for return to a RUN 
             * state.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (Config.sts < RUN_STATE_FAULT)
            {
                //Recover from low battery voltage
                if ((Config.sts == STS_BATL) && \
                        (Config.bat.volt > S_VL_BAT_RSRT))
                {
                    if ((++count_batl) > S_LM_BATL_HYS)
                    {
                        Config.sts = STS_BOOT;
                        break;
                    }
                }
                else if (count_batl) --count_batl;

                //Recover from low PV voltage
                if ((Config.sts == STS_PVVL) && \
                        (Config.pvv.volt > (ramData[FLASH_AR_PVVL_RSRT])))
                {
                    if ((++count_pvvl) > S_LM_PVVL_HYS)
                    {
                        Config.sts = STS_BOOT;
                        break;
                    }
                }
                else if (count_pvvl) --count_pvvl;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Unrecoverable FAULT states: unit is halted.  Make sure Vdd is
             * sufficient and attempt write to NVM.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (ramData[FLASH_AR_STS] != Config.sts)
            {
                ReadVoltage(ADC_CH_DAC_3, S_RT_VDD, &Config.vdd);
                if (Config.vdd.volt > S_LM_VDD_NVM)
                {
                    ramData[FLASH_AR_STS] = Config.sts;
                    writeFlashRow((uint16_t) nvData, ramData);
                }
            }

            break;                              //Always exit @ bottom of loop
        }
#endif

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * SHUT-DOWN.
         * 
         * This is an extension of the FAULT MONITORING block and stops the 
         * charger if running.  This is tested using the macro cogStatus and 
         * evaluating the status code.
         * 
         * Macro getSnapshot takes a snapshot of the system status that 
         * triggered the shutdown for console retrieval.
         * 
         * The hysteresis counters for non-FAULT conditions are cleared for 
         * proper operation of the fault recovery logic.  Set restart timer for 
         * voltages to settle during ongoing sampling to ensure a [more likely] 
         * successful restart and to reduce [possible] oscillation between 
         * operating states.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((cogStatus()) && (Config.sts != STS_RUN))
        {
            ChargerOFF();
            getSnapshot();
            count_batl = 0;
            count_pvvl = 0;
            RestartTimer = S_TM_RESTART;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * RUN MONITORING & CONTROL.
         * 
         * This block controls the unit's operation when running (status code = 
         * RUN):
         *  + Buck low-side synchronous gate drive on-time
         *  + Output SSR relay
         *  + MPPT algorithm
         * 
         * Low-side drive timing is autonomous whenever the unit is running.  
         * The other functions are only automatic when running in closed 
         * loop.  In open loop the relay can be manually controlled via console
         * commands (careful, there are no safety interlocks).  The MPPT 
         * algorithm cannot; it executes automatically in closed loop.
         * 
         * Low-side synchronous gate drive control:
         * Because the HW design lacks circuitry to directly sense freewheeling 
         * current and control the low-side MOSFET, the firmware uses fixed 
         * timing to balance proper operation with efficiency.  If the low
         * side is left on too long at light loads (the inductor is 
         * discontinuous) then the buck core becomes a boost in reverse by
         * charging the inductor with energy from the battery and discharging
         * thru the high-side MOSFET into the PV panels.  Conversely, if the low 
         * side MOSFET remains off, then freewheeling current conducts thru the 
         * body diode resulting in potentially significant heat dissipation (and 
         * potential failure).
         * 
         * Below the PV current threshold of S_LM_PVC_LS, the low side is on for
         * a few hundreds of nano-seconds: long enough to maintain the bootstrap
         * capacitor charge but short enough to prevent reverse operation.  
         * Above the threshold the low side is allowed to remain on for the
         * time modeled at maximum input voltage and 100% load.  In practice,
         * this on-time will be less and is governed by the duty cycle 
         * established by the HW & MPPT control loops.  Note that the threshold
         * is set by integer amps (pvc.volt) so it is one amp higher than 
         * specified by S_LM_PVC_LS.
         * 
         * The low-side gate drive timing uses a rising slope PRG, comparator,
         * and DAC voltage reference to establish the timing.  The COG's low-
         * side signal connects to a CLC configured as an AND gate with the
         * second input connected to the comparator's output.  Hence, the low-
         * side timing is governed by the shorter of the COG & comparator's 
         * outputs.  Additionally, this method allows the COG's dead-band &
         * blanking features to be used.  Refer to the peripheral flowchart for
         * more detail.
         * 
         * Output SSR relay control:
         * A disadvantage of a non-isolated buck is that the output provides a 
         * conduction path back to the input.  In this application / design it
         * isn't necessarily harmful since the low-side on-time won't be 
         * sufficient to create a boost reversal and the PV panel will be 
         * reverse biased.  However, it is inefficient and would prevent the low
         * PV voltage feature from stopping the unit to conserve battery energy.
         * 
         * In closed loop the PV current is monitored: any positive current sets
         * the relay on.  During no-load or, when lighting conditions are 
         * insufficient for the unit to deliver a positive output, the relay is
         * set off.  This allows the fault monitoring block to promptly detect
         * a low PV voltage condition and stop the unit to conserve energy.
         * 
         * MPPT algorithm:
         * Any power supply connected to a source impedance that is variable
         * and often higher than its own must modify its operation to match
         * impedances as closely as possible for maximum efficiency.  This
         * design uses duty cycle for impedance matching and works as follows 
         * (refer to the peripheral config schematic for details).
         * 
         * The hardware control loop, consisting of an inner peak current and 
         * outer voltage loops, runs without interference from the firmware.
         * The comparator output from that loop drives a falling event to the 
         * COG that sets duty cycle.  The maximum duty cycle of the unit is 
         * governed by the master PWM and also generates a falling event to the 
         * COG.  The COG works on a first come, first serve, basis:  the first 
         * falling event ends the DC (subsequent events are ignored).  Under 
         * normal circumstances (without MPPT & low source impedance), the HW 
         * control loop would always generate the falling event.
         * 
         * The MPPT firmware manipulates the maximum allowable DC by modifying 
         * the master PWM's DC.  When the PV output is low (high impedance), it
         * will set a low maximum DC.  Regardless of what the HW control loop
         * demands the master PWM will terminate the DC first.  During periods
         * of high PV production (low impedance) and/or loads less than 
         * available energy, the master PWM DC will be set high and the HW
         * control loop will terminate the DC.  Of course the side effect of 
         * this scheme (or any that manipulates the power supply's impedance) is 
         * that the output voltage will often be less than what the HW control 
         * loop demands.
         * 
         * Two algorithms are included as compile options: hysteresis and sweep.
         * Define constant _MPPT_HEURISTIC accordingly to select.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts == STS_RUN)
        {
#ifdef _TIMING_RUN
            //To measure run loop time
            PIN_OD_RUN ^= TRUE;
#endif

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * LOW-SIDE ON-TIME CONTROL.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (Config.pvc.volt > S_LM_PVC_LS) DAC4REF = CFG_DAC_4_R_HI;
            else DAC4REF = CFG_DAC_4_R_LO;


            if (ramData[FLASH_AR_LOOP])
            {
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * OUTPUT SSR RELAY CONTROL.
                 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (Config.pvc.vmv) setRelayOn();
                else setRelayOff();


                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * MPPT P&O HYSTERESIS ALGORITHM.  The algorithm consists of two 
                 * sections.  The first moves the duty cycle a fixed increment, 
                 * defined by S_TM_MPPT_DC, in the direction indicated by the 
                 * bit flag mpptIncrease when the bit flag mpptMove is set TRUE.
                 * Evaluation of the results of moving the DC is handled by the 
                 * second section.
                 * 
                 * When the DC is moved there is a mandatory delay defined by 
                 * constant S_LM_MPPT_COUNT which specifies the number of RUN 
                 * loop iterations that must complete.  This is done to reduce 
                 * jitter and 'hunting'.  Because it uses loop iterations and 
                 * not a fixed timing source it's important to know the average 
                 * loop execution time since this forms a primitive filter.  
                 * This can be measured by uncommenting the constant _TIMING_RUN
                 * in mpptproto.h and probing one of the LED pins.
                 * 
                 * When the timer expires the current PV power is compared to 
                 * the MPPT's previously sampled power.  If less the DC 
                 * direction is reversed since it's assumed that the peak was 
                 * passed.  If equal, no change is made and a new delay/sample 
                 * taken.  This method creates a sweeping motion over the 
                 * available PV power range.
                 * 
                 * Even in static conditions the algorithm hunts.  Adding 
                 * hysteresis (similar to the fault monitoring method) would 
                 * likely help.
                 * 
                 * REFINEMENTS
                 * v1.1, b6:
                 *      + count_mppt changed to 16-bit integer to allow longer 
                 *        sampling delays.
                 *      + S_LM_MPPT_COUNT changed from (S_LM_SAMPLES + 200) to
                 *        (S_LM_SAMPLES + 580); approximately 1S
                 *      + Added zero-power block to (Config.pwr == 
                 *        Config.mppt_pwr) block.  Addresses low-light and 
                 *        resulting zero-power condition that prevents scanning.
                 *        Forces scanning up & down the DC range.
                 * v1.1, b8:
                 *      + Changed boot entry point in function ChargerOn to 
                 *        delay block since previously it would immediately
                 *        move to next DC increment thereby skipping the first
                 *        one.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _MPPT_HEURISTIC
                if (mpptMove)
                {
                    if (mpptIncrease)
                    {
                        if ((Config.mppt_dc = (uint8_t) (PWM5DCL + \
                                S_TM_MPPT_DC)) > ramData[FLASH_AR_DC])
                                Config.mppt_dc = ramData[FLASH_AR_DC];
                    }
                    else
                    {
                        if ((Config.mppt_dc = (uint8_t) (PWM5DCL - \
                                S_TM_MPPT_DC)) < S_DC_MIN_MPPT)
                                Config.mppt_dc = S_DC_MIN_MPPT;
                    }
                    chgDutyCycle((uint16_t) PWM5DCL, (uint16_t) Config.mppt_dc);
                    mpptMove = FALSE;
                    count_mppt = S_IN_MPPT;
                }
                else if (++count_mppt > S_LM_MPPT_COUNT)
                {
                    mpptMove = TRUE;
                    if (Config.pwr < Config.mppt_pwr) mpptIncrease ^= TRUE;
                    else if (Config.pwr == Config.mppt_pwr)
                    {
                        //ZERO POWER BLOCK:
                        if (!(Config.pwr))
                        {
                            if (mpptIncrease)
                            {
                                if (Config.mppt_dc == ramData[FLASH_AR_DC])
                                    mpptIncrease ^= TRUE;
                            }
                            else if (Config.mppt_dc == S_DC_MIN_MPPT)
                                mpptIncrease ^= TRUE;
                        }
                        else
                        {
                            count_mppt = S_IN_MPPT;
                            mpptMove = FALSE;
                        }
                    }
                    Config.mppt_pwr = Config.pwr;
                }
#else
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * MPPT P&O SWEEP ALGORITHM.  The sweep algorithm periodically
                 * sweeps the DC from minimum to maximum to find MPP, sets it,
                 * sleeps for a defined period, and then evaluates if a new
                 * sweep is required.  The goal of sweep is to reduce the
                 * hunting, variability, and lag in finding MPP inherent in the 
                 * heuretic version.
                 * 
                 * This version uses the same object set and requires the
                 * additional variables:
                 *      + Config.mppt_pdc saves the DC found at peak power 
                 *        during sweep.
                 *      + lock_mppt counts the number of iterative sleep cycles
                 *        to facilitate limiting to prevent a false MPP lock.
                 *      + counter is a function private scratch object used
                 *        to evaluate the current power tolerance at the end of
                 *        a sleep cycle.
                 * 
                 * The mpptMove flag controls the algorithm's execution state.
                 * When TRUE a sweep is performed.  The sweep begins at minimum
                 * DC and ends when the peak power point is past (determined by
                 * a drop in current to saved power) or when peak DC is reached.
                 * It handles zero-power conditions and will default to minimum
                 * DC if the power curve is flat.
                 * 
                 * Object count_mppt times the sweep and sleep sections.  
                 * Constant S_LM_MPPT_SWEEP defines the number of RUN loop 
                 * iterations between sweep increments and should be set to the
                 * equivalent of some hundreds of mS, between 200-500mS.  In the
                 * sleep section the constant S_LM_MPPT_COUNT defines the 
                 * 'sleep' time (again, by RUN loop iterations) and should be 
                 * set to the equivalent of 10's of seconds - 20 ~ 30.  Shorter
                 * sleeps will result in more scans with a potential for lower
                 * long-term power delivery averages but more responsive to
                 * changing load & PV energy conditions; a longer interval gives
                 * the inverse.
                 * 
                 * When the sleep timer expires the current PV power is compared
                 * to the PV power from the time MPP was set.  If within a 
                 * tolerance of S_LM_MPPT_PTOL watts, no sweep is performed and
                 * another sleep begins.  A sweep will always be performed if
                 * the current power is zero to avoid a 'lock' condition.  The
                 * intent of the tolerance factor is to provide more consistent 
                 * power delivery during relatively static periods.
                 * 
                 * Even with this logic there can be situations where a false
                 * MPP lock can occur.  Example - low ambient with a high load
                 * results in a low DC.  The load remains constant but ambient
                 * increases as the sun shines directly on the panel.  With a
                 * low DC the power change may not exceed the tolerance and the
                 * MPP remains falsely locked at a low value.  To prevent this,
                 * object lock_mppt is used to limit the number of iterative
                 * sleep cycles.  The limit is set by constant S_LM_SWEEP_LOCK.
                 * When the limit is reach a new sweep cycle is started.  The
                 * limit value set is a compromise: too high and power may be
                 * lost in variable conditions.  Too low and long-term averages
                 * will be lower due to sweep action.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (mpptMove)                                       //Sweep
                {
                    if (++count_mppt > S_LM_MPPT_SWEEP)
                    {
                        count_mppt = S_IN_MPPT;                     //Reset
                        if (Config.pwr > Config.mppt_pwr)           //Track MPP
                        {
                            Config.mppt_pwr = Config.pwr;
                            Config.mppt_pdc = PWM5DCL;
                        }
                        if (Config.mppt_dc == ramData[FLASH_AR_DC]) //Set MPP
                        {
                            mpptMove = FALSE;
                            lock_mppt = S_IN_SWEEP_LOCK;            //Clear lock
                            Config.mppt_dc = Config.mppt_pdc;       //Clean data
                            chgDutyCycle((uint16_t) PWM5DCL, \
                                    (uint16_t) Config.mppt_pdc);
                        }
                        else                                        //Next incr.
                        {
                            if ((Config.mppt_dc = (uint8_t) (PWM5DCL + \
                                    S_TM_MPPT_DC)) > ramData[FLASH_AR_DC])
                                Config.mppt_dc = ramData[FLASH_AR_DC];
                            chgDutyCycle((uint16_t) PWM5DCL, \
                                    (uint16_t) Config.mppt_dc);
                        }
                    }
                }
                else if (++count_mppt > S_LM_MPPT_COUNT)            //Sleep
                {
                    ++lock_mppt;                                    
                    count_mppt = S_IN_MPPT;
                    if (Config.pwr > Config.mppt_pwr)               //Calc tol.
                        counter = (uint8_t) (Config.pwr - Config.mppt_pwr);
                    else counter = (uint8_t) (Config.mppt_pwr - Config.pwr);
                    if ((counter > S_LM_MPPT_PTOL) || (!(Config.pwr)) || \
                            (lock_mppt > S_LM_SWEEP_LOCK))
                    {                                               //Do sweep
                        mpptMove = TRUE;
                        Config.mppt_pwr = 0;
                        Config.mppt_dc = S_DC_MIN_MPPT;
                        Config.mppt_pdc = Config.mppt_dc;
                        chgDutyCycle((uint16_t) PWM5DCL, \
                                (uint16_t) Config.mppt_dc);
                    }
                }
#endif
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * BOOT (AUTOSTART).
         * 
         * This block is responsible for completing the boot process or resuming
         * a RUN state (from IDLE).
         * 
         * The boot process does not proceed until valid voltage samples are
         * available (refer to data collection block).  RestartTimer must also 
         * be zero.  It is set by the shut-down block to ensure some time for 
         * voltages to dissipate and prevent an oscillatory start/stop condition 
         * when marginal voltages, such as at twilight, might exist.
         * 
         * The hysteresis counters are cleared for proper operation of the fault
         * monitoring logic.  This includes FAULT conditions since they may have
         * been non-zero (but not the cause of shutdown event) from the last
         * RUN state.  Macro getSnapshot takes a snapshot of the system status 
         * that triggered the boot for console retrieval.
         * 
         * If autostart is set (FLASH_AR_AUTO) the charger is started.  Unlike
         * earlier builds the low output voltage check has been removed since
         * low output voltage conditions can occur during startup and MPP
         * scanning and caused false positives during ice & snow removal
         * conditions.  It has been replaced by an over-current check in the
         * FAULT MONITORING & RECOVERY CONTROL block.
         * 
         * If autostart is disabled this code block still executes to complete
         * the boot process and sets the STOP status code.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((Config.sts == STS_BOOT) && (!(RestartTimer)))
        {
            getSnapshot();
            if (ramData[FLASH_AR_AUTO]) ChargerON();
            else Config.sts = STS_STOP;
            count_batl = 0;
            count_pvvl = 0;
            count_ovcc = 0;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * CONSOLE SERVICE.
         * 
         * IMPORTANT:  do not place any code expected to execute on each RUN 
         * loop iteration beyond this point.  The console service handler should 
         * always be placed at bottom of run loop.
         * 
         * This block services the user console accessed via the RS232 micro-USB
         * service port.  The interface is styled after GPIB instruments.  All 
         * commands consist of a two-byte NL (new line) terminated string and 
         * are categorized as get, set, or action commands.  Refer to the 
         * definition of object CmdList in mpptproto.h for command strings,
         * explanation, and values.
         * 
         * For efficiency the command line is parsed so that the command and
         * parameter are available as strings and integer (parameter only).  All
         * commands take no more than one parameter.
         * 
         * Some of the 'Get' commands support retrieval of system values from
         * the snapshot object.  This object is of the same type as Config but
         * only contains valid data for system values.  Its purpose is to 
         * capture the state of the system when an event occurred to be used for
         * troubleshooting or experimentation.  The shut-down & boot blocks 
         * capture a system-generated snapshot.  The console also contains a 
         * command to capture a snapshot.  The system bit flag flag_ss governs 
         * the capture of a snapshot:  if set the system nor console will 
         * overwrite the contents of the snapshot object.  This is done to 
         * preserve the contents for diagnostics.  The flag can only be cleared 
         * with a console command or reboot.  Note that a console initiated 
         * snapshot does not set the flag and can be over-written by a system 
         * event.
         * 
         * The 'get data payload' command is evaluated first since during non-
         * diagnostic operation the LCD will be connected and this shortens the 
         * command evaluation time.  This command sends only 'live' data from
         * Config and not the snapshot since the runtime display philosophy is
         * to display what's happening at the moment polled.  The snapshot is
         * less relevant anyway since all data is sent in one pass versus the
         * GPIB polling in which 100's of RUN loops may occur from start to 
         * finish.
         * 
         * All commands are evaluated in the following if/else block.  Commands 
         * that don't return a value respond with the string 'Ack' (success) or
         * 'Nck' (fail).  Character echo is off by default.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

        //Retrieve command buffer & parse
        if (usartStatRx() != USARTSTSRXRDY) continue;
        cmdb = usartGetRx();                    //Get command buffer
        cmdp = cmdb;
        rmvSpaces(cmdp);                        //Command parameter as string
        cmdi = (unsigned int) atoi(cmdp);       //Parameter as unsigned int
        if (cmdi) ss_ptr = &snapshot;           //Get snapshot data
        else ss_ptr = &Config;                  //Get real-time data
        cmd_ack = NULL;                         //Clear command result message

        //Get data payload (for LCD)
        if (!(stricmp(cmdb, CmdList[CMDGETDAT])))
        {
            //Assign data
            lcd.sts = Config.sts;
            lcd.pwr = Config.pwr;
            lcd.bat_v = Config.bat.volt;
            lcd.bat_mv = Config.bat.mv;
            lcd.pv_v = Config.pvv.volt;
            lcd.pv_mv = Config.pvv.mv;
            lcd.pv_a = Config.pvc.volt;
            lcd.pv_ma = Config.pvc.mv;

            //Transmit without error check
            usartWrite((void *) &lcd, (uint8_t) sizeof(lcd));
        }
        //Get PV amps
        else if (!(strnicmp(cmdb, CmdList[CMDGETAMP], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->pvc.volt, ss_ptr->pvc.mv);
        //Get battery voltage
        else if (!(strnicmp(cmdb, CmdList[CMDGETBAT], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->bat.volt, ss_ptr->bat.mv);
        //Get over-current count
        else if (!(strnicmp(cmdb, CmdList[CMDGETCNT], CMDSIZSTR)))
            printf(fmtGpib, (unsigned int) ss_ptr->ovcc, (unsigned int) 0);
        //Get set duty cycle
        else if (!(stricmp(cmdb, CmdList[CMDGETDCY])))
            printf(fmtGpib, \
                    (unsigned int) ((ramData[FLASH_AR_DC] * 100) / S_TM_MOD_F),\
                    (unsigned int) 0);
        //Get set output volts
        else if (!(stricmp(cmdb, CmdList[CMDGETOVS])))
            printf(fmtGpib, (unsigned int) DAC1REF, (unsigned int) 0);
        //Get charger ofF @ PV voltage
        else if (!(stricmp(cmdb, CmdList[CMDGETPVL])))
            printf(fmtGpib, (unsigned int) ramData[FLASH_AR_PVVL_STOP], \
                    (unsigned int) 0);
        //Get charger Go @ PV voltage
        else if (!(stricmp(cmdb, CmdList[CMDGETPVG])))
            printf(fmtGpib, (unsigned int) ramData[FLASH_AR_PVVL_RSRT], \
                    (unsigned int) 0);
        //Get snapshot of dynamic parameters (ticks could be corrupt)
        else if (!(strnicmp(cmdb, CmdList[CMDGETSNP], CMDSIZSTR)))
        {
            if (flag_ss) cmd_ack = msgNck;
            else
            {
                memcpy((void *) &snapshot, (void *) &Config, sizeof(syscf));
                cmd_ack = msgAck;
            }
        }
        //Get over-current limit
        else if (!(stricmp(cmdb, CmdList[CMDGETCUR])))
            printf(fmtGpib, (unsigned int) DAC2REF, (unsigned int) 0);
        //Get loop setting
        else if (!(stricmp(cmdb, CmdList[CMDGETLPS])))
            printf(fmtGpib, (unsigned int) ramData[FLASH_AR_LOOP], \
                    (unsigned int) 0);
        //Get MPPT operating duty cycle
        else if (!(strnicmp(cmdb, CmdList[CMDGETMDC], CMDSIZSTR)))
            printf(fmtGpib, \
                    (unsigned int) ((ss_ptr->mppt_dc * 100) / S_TM_MOD_F),\
                    (unsigned int) 0);
        //Get NVM status code
        else if (!(stricmp(cmdb, CmdList[CMDGETNVM])))
            printf("%s %X %X\n", StsList[ramData[FLASH_AR_STS]], \
                    (unsigned int) ramData[FLASH_AR_STATUS], \
                    (unsigned int) ramData[FLASH_AR_PCON]);
        //Get output voltage
        else if (!(strnicmp(cmdb, CmdList[CMDGETOUT], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->out.volt, ss_ptr->out.mv);
        //Get firmware release
        else if (!(stricmp(cmdb, CmdList[CMDGETREL])))
            printf(fmtPrg, VERMAJOR, VERMINOR, VERBUILD);
        //Get system status
        else if (!(strnicmp(cmdb, CmdList[CMDGETSTS], CMDSIZSTR)))
            printf("%s\n", StsList[ss_ptr->sts]);
        //Get autostart setting
        else if (!(stricmp(cmdb, CmdList[CMDGETAUT])))
            printf(fmtGpib, (unsigned int) ramData[FLASH_AR_AUTO], \
                    (unsigned int) 0);
        //Get PV voltage
        else if (!(strnicmp(cmdb, CmdList[CMDGETVOL], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->pvv.volt, ss_ptr->pvv.mv);
        //Get PV watts
        else if (!(strnicmp(cmdb, CmdList[CMDGETWAT], CMDSIZSTR)))
            printf(fmtGpib, (unsigned int) ss_ptr->pwr, (unsigned int) 0);
        //Get relay status
        else if (!(stricmp(cmdb, CmdList[CMDGETRLY])))
            printf(fmtGpib, (unsigned int) relayStatus(), (unsigned int) 0);
        //Clear [current] counter
        else if (!(stricmp(cmdb, CmdList[CMDCLRCNT])))
        {
            Config.ovcc = 0;
            cmd_ack = msgAck;
        }
        //Clear NVM error
        else if (!(stricmp(cmdb, CmdList[CMDCLRNVM])))
        {
            ramData[FLASH_AR_STS] = STS_UND;
            ramData[FLASH_AR_STATUS] = (unsigned char) NULL;
            ramData[FLASH_AR_PCON] = (unsigned char) NULL;
            writeFlashRow((uint16_t) nvData, ramData);
            cmd_ack = msgAck;
        }
        //Clear snapshot flag
        else if (!(stricmp(cmdb, CmdList[CMDCLRSNP])))
        {
            flag_ss = FALSE;
            cmd_ack = msgAck;
        }
        //Restore NVM defaults
        else if (!(stricmp(cmdb, CmdList[CMDRSTDFL])))
        {
            writeFlashRow((uint16_t) nvData, NULL);
            cmd_ack = msgAck;
        }
        //Reset MCU
        else if (!(stricmp(cmdb, CmdList[CMDRSTMCU])))
        {
            //Reset system; first do clean stop if running
            if (cogStatus()) ChargerOFF();
            SWDTEN = FALSE;             //Disable so watchdog doesn't erroneously reboot
            PCON = S_LM_PCON_BT;        //Clear PCON bits to clear error trap on boot
            __delay_ms(S_TM_RST);
            RESET();
        }
        //Set autostart
        else if (!(strnicmp(cmdb, CmdList[CMDSETAUT], CMDSIZSTR)))
        {
            if (cmdi) ramData[FLASH_AR_AUTO] = TRUE;
            else ramData[FLASH_AR_AUTO] = FALSE;
            writeFlashRow((uint16_t) nvData, ramData);
            cmd_ack = msgAck;
        }
        //Set per-cycle current limit
        else if (!(strnicmp(cmdb, CmdList[CMDSETCUR], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set the current limit by updating the DAC with user-supplied 
             * value.  The entered value is bounds checked.  When changed the 
             * updated value is stored in NVM.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            if ((cmdi < S_LM_CLLOW) || (cmdi > S_LM_CLHIG)) cmd_ack = msgNck;
            else if (cmdi != *(RDword + FLASH_AR_CLIM))
            {
                *(RDword + FLASH_AR_CLIM) = cmdi;
                writeFlashRow((uint16_t) nvData, ramData);
                DAC2REFH = ramData[FLASH_AR_CLIM_H];
                DAC2REFL = ramData[FLASH_AR_CLIM_L];
                DAC2LD = TRUE;
                while (DAC2LD);
            }
        }
        //Set duty cycle
        else if (!(strnicmp(cmdb, CmdList[CMDSETDCY], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Convert user-entered duty cycle percentage to the PWM register 
             * value, bound check, and store in NVM if updated.  If running, the
             * DC is changed using chgDutyCycle to avoid a transient.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            cmdi = (((cmdi * CFG_PWM_5_PERIOD) / S_RT_PERCENT) + 1);
            if ((cmdi < S_DC_MIN) || (cmdi > S_DC_MAX)) cmd_ack = msgNck;
            else if (cmdi != ramData[FLASH_AR_DC])
            {
                //Update NVM
                ramData[FLASH_AR_DC] = (unsigned char) cmdi;
                writeFlashRow((uint16_t) nvData, ramData);
                //If operating move to new DC now
                if (!(COG1ASD0bits.ASE))
                    chgDutyCycle((uint16_t) PWM5DC, cmdi);
            }
        }
        //Set console echo on / off.  Anything other than 0 turns on.
        else if (!(strnicmp(cmdb, CmdList[CMDSETECH], CMDSIZSTR)))
        {
            if (cmdi) usartSetEcho(TRUE);
            else usartSetEcho(FALSE);
            cmd_ack = msgAck;
        }
        //Set charger ofF @ PV voltage
        else if (!(strnicmp(cmdb, CmdList[CMDSETPVL], CMDSIZSTR)))
        {
            cmd_ack = msgAck;
            if ((cmdi < S_LM_PVV_L) || (cmdi > S_LM_PVV_H)) cmd_ack = msgNck;
            else if (cmdi != (unsigned int) ramData[FLASH_AR_PVVL_STOP])
            {
                ramData[FLASH_AR_PVVL_STOP] = (unsigned char) cmdi;
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }
        //Set charger Go @ PV voltage
        else if (!(strnicmp(cmdb, CmdList[CMDSETPVG], CMDSIZSTR)))
        {
            cmd_ack = msgAck;
            if ((cmdi < S_LM_PVV_L) || (cmdi > S_LM_PVV_H)) cmd_ack = msgNck;
            else if (cmdi != (unsigned int) ramData[FLASH_AR_PVVL_RSRT])
            {
                ramData[FLASH_AR_PVVL_RSRT] = (unsigned char) cmdi;
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }
        //Set loop open / closed
        else if (!(strnicmp(cmdb, CmdList[CMDSETLOP], CMDSIZSTR)))
        {
            if (cogStatus()) cmd_ack = msgNck;          //Must be off
            else
            {
                if (cmdi)                               //Closed
                {
                    ramData[FLASH_AR_LOOP] = TRUE;
                    COG1FIS0 = CFG_COG_1_FIS0;
                }
                else                                    //Open
                {
                    ramData[FLASH_AR_LOOP] = FALSE;
                    COG1FIS0 = CFG_COG_1_FIS0_O;
                }
                if (ramData[FLASH_AR_LOOP] != nvData[FLASH_AR_LOOP])
                    writeFlashRow((uint16_t) nvData, ramData);
                cmd_ack = msgAck;
            }
        }
        //Set charger on / off
        else if (!(strnicmp(cmdb, CmdList[CMDSETCON], CMDSIZSTR)))
        {
            if (cmdi) ChargerON();
            else ChargerOFF();
            cmd_ack = msgAck;
        }
        //Set output voltage
        else if (!(strnicmp(cmdb, CmdList[CMDSETVOL], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set the output voltage by updating the DAC with user-supplied 
             * value.  The entered value is bounds checked.  When changed the 
             * updated value is stored in NVM.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            if ((cmdi < S_LM_VLOW) || (cmdi > S_LM_VHIG)) cmd_ack = msgNck;
            else if (cmdi != *(RDword + FLASH_AR_VSET))
            {
                *(RDword + FLASH_AR_VSET) = cmdi;
                DAC1REFH = ramData[FLASH_AR_VSET_H];
                DAC1REFL = ramData[FLASH_AR_VSET_L];
                DAC1LD = TRUE;
                while (DAC1LD);
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }
        //Set relay on / off
        else if (!(strnicmp(cmdb, CmdList[CMDSETRLY], CMDSIZSTR)))
        {
            if (cmdi) setRelayOn();
            else setRelayOff();
            cmd_ack = msgAck;
        }

        if (cmd_ack) printf(cmd_ack);   //Print command result
        usartSetRx(ENABLE);             //Look for new user input
    }
} //End function main


/* Function ChargerON.  Starts charger; if already started it has no effect.
 * No pre-start diagnostic voltage checks are done since this is expected to be
 * handled by the RUN loop and reflected in the value of object Config.sts.
 * 
 * If operating in closed loop the DC is increased to the MPPT minimum.  In
 * open loop the duty cycle is set to the NVM value.  DAC4's voltage is set
 * to the value for minimum low-side on time; the run loop governs its increase
 * based on load.
 * 
 * The MPPT bit flags are set TRUE and the MPPT power & duty cycle values are
 * cleared regardless of the loop setting.
 * 
 * Returns:  TRUE if started; Config.sts set to RUN.
 *           FALSE if not; Config.sts contains error code.
 *
 * Parameters:  none
 *
*/
uint8_t ChargerON()
{
    uint8_t result;

    result = FALSE;

    if (COG1ASD0bits.ASE && ((Config.sts == STS_STOP) || \
            (Config.sts == STS_BOOT)))
    {
        DAC4REF = CFG_DAC_4_R_LO;               //Set low-side to min. on-time
        PRG1CON0bits.GO = TRUE;                 //Start slope compensation ramp
        PRG2CON0bits.GO = TRUE;                 //Start low-side timing ramp
        PWM5DC = S_DC_MIN;                      //Soft start with min DC
        PWM5LDCONbits.LDA = TRUE;
        COG1ASD0bits.ASE = FALSE;               //Output on

        if (ramData[FLASH_AR_LOOP])             //Closed loop
            chgDutyCycle(S_DC_MIN, (uint16_t) S_DC_MIN_MPPT);
        else                                    //Open loop
            chgDutyCycle(S_DC_MIN, (uint16_t) ramData[FLASH_AR_DC]);

        Config.sts = STS_RUN;
        Config.ovcc = 0;
        Config.mppt_pwr = 0;
        Config.mppt_dc = S_DC_MIN_MPPT;
        Config.mppt_pdc = Config.mppt_dc;
        mpptIncrease = TRUE;
        count_mppt = S_IN_MPPT;

#ifdef _MPPT_HEURISTIC
        mpptMove = FALSE;
#else
        mpptMove = TRUE;
#endif

        setledRUN();
        result = TRUE;
    }

    return(result);
} //End function ChargerON


/* Function ChargerOFF.  Stops charger by disabling the COG.  The output relay
 * is first turned off to prevent a discharge condition thru the PV panel via
 * the buck high-side switch.  Both ramps (PRGs) are stopped to reduce bias
 * current and switching noise when idle.
 * 
 * The status is set to STOP only if was set to RUN to preserve the condition
 * that initiated the STOP.
 * 
 * Returns:  none.  Config.sts is set to STOP provided it doesn't contain an 
 *           error code.
 *
 * Parameters:  none
 *
*/
void ChargerOFF()
{
    setRelayOff();
    COG1ASD0bits.ASE = TRUE;
    PRG1CON0bits.GO = FALSE;                    //Stop slope compensation ramp
    PRG2CON0bits.GO = FALSE;                    //Stop low-side timing ramp
    if (Config.sts == STS_RUN) Config.sts = STS_STOP;
    setledFAULT();

    return;
} //End function ChargerOFF


/* Function ReadVoltage.  Reads the value from the analog pin channel, 
 * interprets as a scaled voltage, and stores in the volts object pointed to by 
 * rv.  The FVR can be set to low range for higher resolution or for measurements
 * where the high range exceeds Vdd.  The default is high range.  The mvr 
 * variable of rv must be set to the proper range.
 * 
 * The integer, decimal, and VMV components are stored in separate elements of 
 * the object pointed to by rv.  All three are stored with the resolution 
 * defined by system constants.  See details in the 'ADC configuration, 
 * acquisition, and value storage' section of mpptproto.h.
 * 
 * No error checking is performed and only a single sample is taken.
 * 
 * Returns:  none.  The following values are populated in object rv:
 *      vmv=composite value, resolution defined by S_RT_VMV_INT & S_RT_VMV_DEC
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=unmodified value read from ADC; not scaled
 *      mvr=mV per ADC bit set (unmodified)
 *
 * Parameters:  channel=analog channel.  The pin must be properly configured.
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *              rv.mvr must be set to S_RT_ADC_MV_H or S_RT_ADC_MV_L
 *
*/
void ReadVoltage(unsigned char channel, unsigned char scale, volts *rv)
{
    static union {
        unsigned char result[S_SZ_ADC_INT];
        unsigned int value;
    } adc;
    static unsigned int adc1v = S_QY_ADC_1V_H;  //ADC value for 1V

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Connect ADC to channel and acquire.  Concurrently assign FVR range to
     * allow for settle time and setup conversion constants.  The ADC mV per bit
     * resolution is stored in the user-supplied volts object for use by other 
     * conversion functions such as CalcAvg and should not be modified.
     * 
     * For the low scale, the constants FI_FVR_DLY and S_TM_ADC_ACQ must ensure 
     * that both the ADC acquisition and FVR settle requirements are met.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (unsigned) (channel | S_EN_ADC);   //Connect to channel & enable ADC
    if (rv->mvr == S_RT_ADC_MV_L)
    {
        FVRCONbits.ADFVR = CFG_FVR_ADC_2X;
        adc1v = S_QY_ADC_1V_L;
        __delay_us(FI_FVR_DLY);                 //Refresh S&H cap, FVR settle
    }
    else __delay_us(S_TM_ADC_ACQ);              //Refresh S&H cap

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Perform conversion, calculate, scale, and store result.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0bits.GO = TRUE;                       //Start conversion
    while (ADCON0bits.GO);                      //Wait for completion
    adc.result[POS_ADC_HB] = ADRESH;
    adc.result[POS_ADC_LB] = ADRESL;
    rv->adc = adc.value;                        //Save raw ADC value to user object
    adc.value *= scale;                         //Scale value
    rv->volt = (adc.value / adc1v);
    rv->mv = (((adc.value % adc1v) * rv->mvr) / S_RT_VMV_DEC);
    rv->vmv = ((rv->volt * S_RT_VMV_INT) + rv->mv);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set FVR back to high range.  Expect that given MCU speed & code overhead
     * sufficient time exists for FVR to settle without accommodation.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (rv->mvr == S_RT_ADC_MV_L)
    {
        FVRCONbits.ADFVR = CFG_FVR_ADC_4X;
        adc1v = S_QY_ADC_1V_H;
    }

    return;
} //End function ReadVoltage


/* Function CalcAvg.  Calculates the arithmetic mean of a sample set of 
 * previously acquired values.  No bounds checking is performed.
 * 
 * The value to be averaged is expected to represent the sum of raw ADC
 * conversions; they should not be scaled.  The average is calculated and then
 * scaled by this function.  
 * 
 * This function stores the result in the object of type volts pointed to by rv.
 * See function ReadVoltage for details.
 * 
 * Returns:  none.  The following values are populated in object rv.  All 
 * represent averaged values:
 *      vmv=composite value, resolution defined by S_RT_VMV_INT & S_RT_VMV_DEC
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=averaged & scaled ADC value
 *      mvr=mV per ADC bit set (unmodified)
 *
 * Parameters:  sum=raw summed ADC values
 *              count=number of summed values
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.  Object mvr in rv must be
 *              initialized to the proper ADC resolution.  See ReadVoltage for
 *              details.
 *
*/
void CalcAvg(unsigned int sum, unsigned char count, unsigned char scale, \
        volts *rv)
{
    static unsigned int adc1v = S_QY_ADC_1V_H;  //ADC value for 1V

    if (rv->mvr == S_RT_ADC_MV_L) adc1v = S_QY_ADC_1V_L;
    else adc1v = S_QY_ADC_1V_H;

    rv->adc = ((sum / count) * scale);
    rv->volt = (rv->adc / adc1v);
    rv->mv = (((rv->adc % adc1v) * rv->mvr) / S_RT_VMV_DEC);
    rv->vmv = ((rv->volt * S_RT_VMV_INT) + rv->mv);

    return;
} //End function CalcAvg


/* Function chgDutyCycle.  Moves the duty cycle from the start to final value 
 * incrementally for smooth operation & transient control.  The constant
 * S_TM_DCDELAY sets the delay, in uS, between DC increments.
 * 
 * Returns:     none
 *
 * Parameters:  start=16-bit starting DC value
 *              final=16-bit final DC value
 *
*/
void chgDutyCycle(uint16_t start, uint16_t final)
{
    while (TRUE)
    {
        PWM5DC = start;
        PWM5LDCONbits.LDA = TRUE;
        if (start == final) break;

        if (start < final) ++start; else --start;
        __delay_us(S_TM_DCDELAY);
    }

    return;
} //End function chgDutyCycle


/* Function writeFlashRow.  Writes a row of flash.  A 16-bit checksum for data
 * is first calculated and stored in the last word of the row.  Hence, the user
 * only has (FLASHROWSZ - 2) bytes available for use.
 * 
 * This function can also be used to restore default NVM settings by leveraging
 * the 'BOOT & NVM PROTECTION AND MANAGEMENT' documented in function
 * Initialize.  To do so this function should be called with the data parameter
 * set to NULL.  When so the checksum word is cleared so that upon the next
 * boot the NVM defaults will be used since NVM data is considered invalid.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit address of row in flash to write.  Must point to
 *                  start of array.
 *              data=pointer to array of data to write to flash -OR-
 *                   NULL to clear the existing checksum but leave other data
 *                   in the row intact.
 *              
 *
*/
void writeFlashRow(uint16_t addr, unsigned char *data)
{
    uint16_t chksum;
    unsigned char count;

    //Writing full row
    if (data)
    {
        count = 0;                              //Start at beginning of row
        chksum = nvmChksum(data);               //Calculate & store checksum
        *(data + FLASH_AR_CSUM_L) = (unsigned char) (chksum & S_MK_U16_L);
        *(data + FLASH_AR_CSUM_H) = \
                (unsigned char) ((chksum & S_MK_U16_H) >> S_SZ_BYTE);

        //Must erase row first for clean write
        eraseFlashRow(addr);
    }
    //Clearing checksum only
    else
    {
        count = FLASH_AR_CSUM_L;                //Start at checksum offset in row
        chksum = 0;                             //Set checksum to zero
        data = (unsigned char *) &chksum;       //Set data address to checksum
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup to write data to flash.  Interrupts must be disabled for entire
     * sequence.  Set LWLO to load data to write latches and then to flash on
     * last byte.
     * 
     * Load high address register here & only once.  Update low register for 
     * individual bytes to write within row.
     * 
     * Variable count represents array element to write.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    di();                               //Interrupts disabled for entire sequence.
    WREN = TRUE;                        //Enable write
    CFGS = FALSE;                       //Program memory
    FREE = FALSE;                       //Write data
    LWLO = TRUE;                        //Load data latches
    PMADRH = (unsigned char) (addr >> 8);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important:  count must be initialized at beginning of function based on 
     * intent.  It controls number of bytes to write and address offset within
     * NVM row.  data must point to full row or checksum only.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    for (; (count < FLASHROWSZ); ++count, ++data)
    {
        PMADRL = (unsigned char) (addr + count);
        PMDATL = *data;
        unlockSeq();                    //Unlock for write
    }

    LWLO = FALSE;                                   //Write row
    unlockSeq();                                    //Write flash

    //All done:  disable writes & re-enable interrupts
    WREN = FALSE;
    ei();

    return;
} //End function writeFlashRow


/* Function eraseFlashRow.  Erases a row of program flash memory.  Interrupts
 * are disabled during the erase.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit pointer to starting address for row.
 *
*/
void eraseFlashRow(uint16_t addr)
{
    PMADRL = (unsigned char) addr;
    PMADRH = (unsigned char) (addr >> S_SZ_BYTE);

    //Erase row
    di();                               //Disable interrupts for duration of function
    CFGS = FALSE;                       //Program memory
    FREE = TRUE;                        //Erase row
    WREN = TRUE;                        //Enable erase
    unlockSeq();                        //Unlock for erase
    WREN = FALSE;                       //Disable erase
    ei();                               //Enable interrupts

    return;
} //End function eraseFlashRow


/* Function nvmChksum.  Calculates the checksum for a row of data.  The 'row'
 * is an unsigned char array of length FLASHROWSZ since this function is
 * intended to be used data storage in NVM (program flash).
 * 
 * This function is intended to be used in conjunction with writeFlashRow and
 * assumes the checksum resides in the last two bytes of the row.  Hence, when
 * calculating the checksum only (FLASHROWSZ - 2) bytes are examined.
 * 
 * The checksum is a simple sum of byte values in the row starting with offset
 * 0.
 * 
 * Returns:  16-bit checksum value
 *
 * Parameters:  data=pointer to data row
 *              
 *
*/
uint16_t nvmChksum(unsigned char *data)
{
    uint16_t sum;
    unsigned char count;

    for (sum = 0, count=0; (count < (FLASHROWSZ - FLASH_SZ_CSUM)); ++count)
        sum += *(data + count);

    return(sum);
} //End function nvmChksum


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts which should be ordered according to priority high-low:
 *      + CM2 over-current counter
 *      + Timer6 (system timer)
 *      + USART Rx (service port RS232 communications)
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    static uint8_t msc = 0;                 //Count of timer6 interrupts

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CM2 ISR HANDLER:  count over-current events.  No accommodation is made
     * for wrapping of counter.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (C2IE & C2IF)
    {
        ++Config.ovcc;
        C2IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SYSTEM TIMER.  Non-precision timed background service.  ~10 mS frequency.
     * Maintains timing objects:
     * + RestartTimer - mS delay before autostart can execute
     * + cal_secs - S delay between PV current calibration
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR6IE & TMR6IF)
    {
        if (RestartTimer) --RestartTimer;
        if (++msc == S_QY_TMR6_1S)
        {
            msc = 0;
            if ((++cal_secs) > S_TM_CAL_PVC) cal_req = TRUE;
        }
        TMR6IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * USART HANDLER:  service RS232 console.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartISR();

    return;
} //End function InterruptExec


/*Function Initialize.  called immediately after boot by function main.  
 * Configures system clock, validates PIC boot reason & sets status, validates
 * NVM & loads to RAM, configures pins & peripherals, and configures interrupts.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization
    Config.sts = STS_BOOT;              //Start with assumption of clean boot...

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * System clock & watchdog config
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON = CFG_OSC_C;                 //Configure system clock
    while (!(HFIOFL));                  //Wait for oscillator to stabilize
    WDTCON = CFG_WDT_CON;               //Watchdog


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * BOOT & NVM PROTECTION AND MANAGEMENT.
     *
     * Recovering from CPU/firmware failures (e.g. watchdog, stack OF/UF, etc.)
     * and NVM corruption is critical for safety and field reliability.  While
     * neither should happen unless there's a true failure, statistically it's 
     * possible when operating in environmental extremes or the more likely
     * existence of a firmware bug.
     * 
     * The goal is to maintain safe operation, keep the unit intact, and record
     * the failure state (in that order).
     * 
     * DATA MANAGEMENT
     * A single row (32 bytes) of NVM data is used for the storage of tunable
     * parameters (e.g. current limit, output voltage, etc.).  Object nvData
     * resides in flash (NVM) and is initialized with default values.  During
     * normal operation object ramData contains a copy of nvData and is used
     * to update nvData.  It must also be initialized with the same default
     * values as nvData.
     * 
     * DATA VALIDATION
     * NVM data is validated with a simple 16-bit checksum stored in the last
     * two bytes of nvData.
     * 
     * BOOT PROCESS
     * The CPU boot status is always checked first to ensure that any abnormal
     * condition is identified before being cleared by subsequent execution.  
     * The PCON & STATUS registers are examined for reboot due to watchdog, 
     * stack OF/UF, and possibly other conditions.  If found the status byte
     * (Config.sts) is updated and it's expected that the main RUN loop halts
     * normal operation.
     * 
     * The NVM checksum is verified before attempting to load.  The checksum
     * value is read directly from NVM and compared to the NVM's calculated
     * checksum.  If they match nvData is copied to ramData.
     * 
     * If there's a mismatch ramData retains it's default values and the status
     * code will be updated if not already set by the CPU checks.
     * 
     * The remainder of the initialization function continues unaltered and
     * when complete the main function enters the RUN loop.  The charger
     * will not start if any status code other than BOOT is set.  A code block 
     * in the main function loop is responsible for storing error codes, 
     * resulting from the boot process, or otherwise, in NVM.
     * 
     * Since the MCU's architecture requires the entire NVM row to be written
     * it's critical that ramData contains valid data.  This is why it's not
     * updated with nvData when checksum's don't match.  By writing the status
     * code it also ensures that if NVM was truly corrupt it will now contain
     * valid data (provided there's not an MCU flash failure) and there's a 
     * chance of the unit returning to operation when power cycled.
     * 
     * Compile note:  braces are intentionally used even though each side of
     * the evaluation contains only one statement.  During testing it was 
     * found that the compiler did not interpret the evaluation correctly 
     * without them.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((PCON & S_LM_PCON_EX) != S_LM_PCON_BT) || \
            ((STATUS & S_LM_STAT_EX) != S_LM_STAT_BT)) Config.sts = STS_MCU;

    if (nvmChksum((unsigned char *) nvData) != (unsigned int) ((((unsigned int) \
            nvData[FLASH_AR_CSUM_H]) << S_SZ_BYTE) | ((unsigned int) \
            nvData[FLASH_AR_CSUM_L])))
    {
        if (Config.sts == STS_BOOT) Config.sts = STS_NVM;
    }
    else
    {
        memcpy((void *) ramData, (void *) nvData, FLASHROWSZ);
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZE CONTROL DATA:
     * + Config.bat.mvr set to high range
     * + Config.pvv.mvr set to high range
     * + Config.pvc.mvr set to high range
     * + Config.out.mvr set to high range
     * + Config.vdd.mvr set to low range
     * + RDword set to ramData array to address NVM data as words for 16-bit
     *   values
     * + Enable individual pin weak pull-ups
     * + Snapshot flag set FALSE (no system event captured)
     * + cal_req is set TRUE to force current sense calibration at start of run
     *   loop.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Config.bat.mvr = S_RT_ADC_MV_H;
    Config.pvv.mvr = Config.bat.mvr;
    Config.pvc.mvr = Config.bat.mvr;
    Config.out.mvr = Config.bat.mvr;
    Config.vdd.mvr = S_RT_ADC_MV_L;
    RDword = (uint16_t *) ramData;
    OPTION_REGbits.nWPUEN = FALSE;
    flag_ss = FALSE;
    cal_req = TRUE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SET PORT CONFIGURATION:
     * + Set digital latch initial states
     * + Pin in/out-put states
     * + Analog / digital
     * + Weak pull-ups
     * + Open drain (tri-state)
     * + Slew rate
     * + Logic input levels
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA = S_IN_PORTA;
    PORTB = S_IN_PORTB;
    PORTC = S_IN_PORTC;
    //Port A
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port B
    TRISB = IO_PORTB;
    ANSELB = AD_PORTB;
    WPUB = WP_PORTB;
    ODCONB = OD_PORTB;
    SLRCONB = SR_PORTB;
    INLVLB = IL_PORTB;
    //Port C
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set the status LED so there is some indication that the unit has power
     * if it hangs before getting to the run loop.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    setledFAULT();


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * MAP PERIPHERAL PINS:
     * + High-side gate drive (COG1)
     * + Synchronous low-side gate drive (CLC1)
     * + SSR (relay) gate drive (PWM3)
     * + PRG2 falling & rising triggers
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_COG1_H = PPS_M_COG1_H;            //High side gate drive
    PPS_P_CLC1_L = PPS_M_CLC1_L;            //Synchronous low side gate drive
    PPS_P_PWM3_SSR = PPS_M_PWM3_SSR;        //PWM3
    PRG2RPPS = PPS_M_PRG2_R;                //Ramp rising trigger
    PRG2FPPS = PPS_M_PRG2_F;                //Ramp falling trigger

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize USART for command console.  This should be done at end of 
     * port / pin config since it sets its respective pins for the USART.
     * 
     * Default is console echo OFF.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartSetRx(ENABLE);
    usartSetEcho(FALSE);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE FVR & DACs:
     * FVR=ADC & DAC Vref
     * 1=Vref for voltage control loop
     * 2=current limit
     * 3=Vdd measurement
     * 4=Synchronous low-side on-time Vref for CM3
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    __delay_us(FI_FVR_DLY);

    DAC1CON0 = CFG_DAC_1_C;
    DAC1REFH = ramData[FLASH_AR_VSET_H];
    DAC1REFL = ramData[FLASH_AR_VSET_L];
    DAC1LD = TRUE;                              //Load reference value

    DAC2CON0 = CFG_DAC_2_C;
    DAC2REFH = ramData[FLASH_AR_CLIM_H];
    DAC2REFL = ramData[FLASH_AR_CLIM_L];
    DAC2LD = TRUE;                              //Load reference value

    DAC3CON0 = CFG_DAC_3_C;
    DAC3REF = CFG_DAC_3_R;

    DAC4CON0 = CFG_DAC_4_C;
    DAC4REF = CFG_DAC_4_R_LO;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE ADC:  enable and connect to DAC3 (Vdd measurement) as default.
     * It's expected that all further management of ADC is done thru function
     * ReadVoltage.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (ADC_CH_DAC_3 | S_EN_ADC);
    ADCON1 = CFG_ADC_CON_1;
    ADCON2 = CFG_ADC_CON_2;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE OP-AMPS:
     * 1=voltage control loop error amp
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA1CON = CFG_OPA_1_C;
    OPA1ORS = CFG_OPA_1_O;
    OPA1NCHS = CFG_OPA_1_N;
    OPA1PCHS = CFG_OPA_1_P;
    OPA1EN = TRUE;                          //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE COMPARATORS:
     * 1=current loop modulator (PWM) control
     * 2=over-current sense
     * 3=synchronous low-side timing
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM1CON0 = CFG_CM_1_C0;
    CM1CON1 = CFG_CM_1_C1;
    CM1NSEL = CFG_CM_1_N;
    CM1PSEL = CFG_CM_1_P;

    CM2CON0 = CFG_CM_2_C0;
    CM2CON1 = CFG_CM_2_C1;
    CM2NSEL = CFG_CM_2_N;
    CM2PSEL = CFG_CM_2_P;

    CM3CON0 = CFG_CM_3_C0;
    CM3CON1 = CFG_CM_3_C1;
    CM3NSEL = CFG_CM_3_N;
    CM3PSEL = CFG_CM_3_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE COG:
     * 1=synchronous buck.  Falling & rising deadband (DBF/DBR) are set from
     * ramData (NVM).  ramData also controls if CM1 is set as a falling event
     * for closed or [not set for] open.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1DBR = ramData[FLASH_AR_RDB];
    COG1DBF = ramData[FLASH_AR_FDB];
    COG1BLKR = CFG_COG_1_BLKR;
    COG1BLKF = CFG_COG_1_BLKF;
    COG1PHR = CFG_COG_1_PHR;
    COG1PHF = CFG_COG_1_PHF;
    COG1ASD1 = CFG_COG_1_ASD1;
    COG1ASD0 = CFG_COG_1_ASD0;
    COG1RIS0 = CFG_COG_1_RIS0;
    COG1RIS1 = CFG_COG_1_RIS1;
    if (ramData[FLASH_AR_LOOP]) COG1FIS0 = CFG_COG_1_FIS0;
    else COG1FIS0 = CFG_COG_1_FIS0_O;
    COG1FIS1 = CFG_COG_1_FIS1;
    COG1RSIM0 = CFG_COG_1_RSIM0;
    COG1RSIM1 = CFG_COG_1_RSIM1;
    COG1FSIM0 = CFG_COG_1_FSIM0;
    COG1FSIM1 = CFG_COG_1_FSIM1;
    COG1CON1 = CFG_COG_1_CON1;
    COG1CON0 = CFG_COG_1_CON0;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE CLCS:
     * 1=low-side synchronous drive
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC1SEL0 = CFG_CLC_1_SEL_0;
    CLC1SEL1 = CFG_CLC_1_SEL_1;
    CLC1SEL2 = CFG_CLC_1_SEL_2;
    CLC1SEL3 = CFG_CLC_1_SEL_3;
    CLC1GLS0 = CFG_CLC_1_GLS_0;
    CLC1GLS1 = CFG_CLC_1_GLS_1;
    CLC1GLS2 = CFG_CLC_1_GLS_2;
    CLC1GLS3 = CFG_CLC_1_GLS_3;
    CLC1POL = CFG_CLC_1_POL;
    CLC1CON = CFG_CLC_1_CON;
    LC1EN = TRUE;                       //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE TIMERS:
     * + Assign timers to PWMs
     * 2=time base for PWM3
     * 6=ISR timing service
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CCPTMRS = CFG_CCP_TMRS;

    T2CLKCON = CFG_TMR_2_CLK;
    T2HLT = CFG_TMR_2_HLT;
    T2RST = CFG_TMR_2_RST;
    T2PR = CFG_TMR_2_PR;
    T2CON = CFG_TMR_2_CON;

    T6CLKCON = CFG_TMR_6_CLK;
    T6HLT = CFG_TMR_6_HLT;
    T6RST = CFG_TMR_6_RST;
    T6PR = CFG_TMR_6_PR;
    T6CON = CFG_TMR_6_CON;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE PWMS:
     * 3=SSR on/off control; PWM is started now but with 0 DC so off
     * 5=master for buck modulator.  The maximum allowable, or open-loop duty
     * cycle, is set by ramData (NVM).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM3DCL = CFG_PWM_3_DCL;
    PWM3DCH = CFG_PWM_3_DCH_OFF;
    PWM3CON = CFG_PWM_3_CON;

    PWM5INTE = CFG_PWM_5_INTE;
    PWM5CLKCON = CFG_PWM_5_CLK;
    PWM5LDCON = CFG_PWM_5_LDCON;
    PWM5OFCON = CFG_PWM_5_OFCON;
    PWM5PH = CFG_PWM_5_PHASE;
    PWM5DC = (uint16_t) ramData[FLASH_AR_DC];
    PWM5PR = CFG_PWM_5_PERIOD;
    PWM5OF = CFG_PWM_5_OFFSET;
    PWM5TMR = CFG_PWM_5_TIMER;
    PWM5CON = CFG_PWM_5_CON;                //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE PRGS:
     * 1=slope compensation.  The ramp slope is set by ramData (NVM).
     * 2=low-side gate drive timing ramp
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PRG1FTSS = CFG_PRG_1_FTSS;
    PRG1RTSS = CFG_PRG_1_RTSS;
    PRG1CON2 = ramData[FLASH_AR_RAMP];
    PRG1INS = CFG_PRG_1_INS;
    PRG1CON1 = CFG_PRG_1_CON1;
    PRG1CON0 = CFG_PRG_1_CON0;

    PRG2FTSS = CFG_PRG_2_FTSS;
    PRG2RTSS = CFG_PRG_2_RTSS;
    PRG2CON2 = CFG_PRG_2_CON2;
    PRG2INS = CFG_PRG_2_INS;
    PRG2CON1 = CFG_PRG_2_CON1;
    PRG2CON0 = CFG_PRG_2_CON0;
    

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZATION COMPLETE:
     * + Enable peripheral & global interrupts
     * + Enable peripheral interrupts serviced by ISR
     * + Start system timer
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts
    C2IE = TRUE;
    TMR6IE = TRUE;
    T6ON = TRUE;

    return;
} //End function Initialize



//End of file