/*

DESCRIPTION:
    Main include file for mpptproto.c - prototype PV charger.

COPYRIGHT:      (c) 2019, Brian Cornell, all rights reserved.

 TARGET MCU:    PIC16F1769

HW REVISION:    1

HW RELEASE:     prototype

VERSION:        1.1

MODIFICATION HISTORY:

    6 Sep 19, BWC, created release.


*/



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Version constants & definitions.  Build number is defined in constant 
 * VERBUILD and maintained by shell script build.sh and included in build.h.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define VERMAJOR    1
#define VERMINOR    1


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Necessary includes.  Build.h contains the build number constant which is
 * maintained by the makefile.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifndef UINT_MAX
#include <limits.h>
#endif
#include "build.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Conditional compilation.  Uncomment directive to enable / disable features.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Uncomment to disable limit checks
//#define _LIMITS_OFF

//Uncomment to check run loop time
//#define _TIMING_RUN

//P&O MPPT algorithm: uncomment to use heuristic -OR- comment out to use sweep
//#define _MPPT_HEURISTIC


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                  Configure system clock & Watchdog Timer
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OSC_C       0x78            //16MHz, FOSC config word determines clock
#define CFG_WDT_CON     0x10            //Watchdog interval=256mS, off.


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * System wide constants & definitions.  Convention:  S_<TYPE>_<INFO>
 * Types:
 *  + DC=Duty Cycle
 *  + DF=Default value for a configurable parameter
 *  + EN=ENable bit
 *  + RT=RaTio or scaling factor
 *  + LM=LiMiting operational value
 *  + TM=TiMing
 *  + MP=Peripheral port bit MaPping
 *  + CH=character
 *  + IN=INitializing value
 *  + VL=VoLtage
 *  + QY=quantity (count) of something
 *  + CD=command data
 *  + BL=boolean value
 *  + MK=bit MasK
 *  + SZ=size of an object or data type
 * 
 * Info is user defined and should be descriptive.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_TM_MOD_F          79              //PWM (modulator) frequency
#define S_DC_MIN            8               //Start & min. open loop DC
#define S_DC_MIN_MPPT       30              //Minimum & starting MPPT DC
#define S_DC_MAX            72              //Maximum allowable duty cycle
#define S_DC_RUN            S_DC_MAX        //Default running duty cycle
#define S_LM_PVV            27              //Maximum expected open PV voltage
#define S_RT_VDD            4               //Vdd ratio rendered by DAC
#define S_RT_PVV            11              //PV volts measurement ratio
#define S_RT_PVC            1               //PV current measurement ratio
#define S_RT_BAT            4               //Battery volts measurement ratio
#define S_RT_OUT            4               //Output volts measurement ratio
#define S_TM_TMR6           10              //Timer6 period in mS
#define S_QY_MS_S           1000            //Number of mS in one second
#define S_QY_TMR6_1S        (S_QY_MS_S / S_TM_TMR6) //# of timer6 periods=1S
#define S_TM_CAL_PVC        64800           //Calibrate current sense every # S
#define S_LM_CAL_PVC        10              //# PV current calibration samples
#define S_QY_MVA_PVC        100             //mV/A of PV current sense circuit
#define S_QY_SAM_PVC        1               //# samples to convert pvc to amps
#define S_RT_MVA_PVC        10              //Scale pvc mV/A to integer amps
#define S_IN_SAMPLES        1               //Starting value for measured samples counter
#define S_IN_MPPT           1               //Starting value for MPP sampling counter
#define S_LM_SAMPLES        5               //Number of measured samples to average
#define S_RT_VDD            4               //Vdd measurement ratio
#define S_LM_VDD_NVM        3               //Vdd > to allow NVM writes
#define S_LM_CLLOW          1               //Lowest allowable current limit value
#define S_LM_CLHIG          1023            //Highest allowable current limit value
#define S_LM_VLOW           690             //Lowest allowable voltage set value
#define S_LM_VHIG           940             //Highest allowable voltage set value
#define S_LM_SLHIG          31              //Highest allowable slope set value
#define S_SZ_BYTE           8               //Number of bits in a byte
#define S_MK_U16_L          0x00FF          //Mask to access low byte of 16-bit unsigned
#define S_MK_U16_H          0xFF00          //Mask to access high byte of 16-bit unsigned
#define S_RT_PERCENT        100             //Used as divisor to calculate % of a dividend value
#define S_TM_DCDELAY        500             //# uS to wait between each DC increment
#define S_TM_MONINTVL       100             //# SysTimer clock ticks for monitor updates (1S)
#define S_TM_RST            200             //# mS to wait before reseting PIC
#define S_TM_INT            300             //# mS for system to stabilize at power-up
#define S_LM_PCON_EX        0xFA            //PCON bits to examine at boot
#define S_LM_PCON_BT        0x18            //PCON expected value at boot
#define S_LM_STAT_EX        0x18            //STATUS bits to examine at boot
#define S_LM_STAT_BT        0x18            //STATUS expected value at boot
#define S_VL_BAT_MIN        10              //Charger won't run below this
#define S_VL_BAT_RSRT       S_VL_BAT_MIN    //Bat must be > for restart
#define S_VL_PV_STOP        13              //PV volts to stop charger
#define S_VL_PV_RSRT        17              //PV volts to start charger
#define S_TM_MPPT_DC        4               //MPPT algorithm DC increment .25uS
#define S_LM_PVC_LS         1               //PV amps > extend low side on-time
#define S_TM_RESTART        250             //# x 10 = mS to delay restart
#define S_LM_PVV_L          12              //Lowest PV volts setting
#define S_LM_PVV_H          25              //Highest PV volts setting
#define S_LM_BATL_HYS       5               //Low battery voltage hysteresis
#define S_LM_PVVL_HYS       5               //Low PV voltage hysteresis
#define S_LM_OVCC_CNT       100             //Max over-current count per loop
#define S_LM_OVCC_HYS       200             //Over-current hysteresis
#define S_CH_SPACE          ' '             //White space
#define S_LM_MPPT_PTOL      5               //Peak PV power tolerance for MPP
#define S_IN_SWEEP_LOCK     1               //Starting value for object lock_mppt
#define S_LM_SWEEP_LOCK     4               //Sleep lock limit for object lock_mppt

#ifdef _MPPT_HEURISTIC
#define S_LM_MPPT_COUNT     (S_LM_SAMPLES + 580)    //# run loops for MPPT P sample
#else
#define S_LM_MPPT_COUNT     (S_LM_SAMPLES + 13500)  //# run loops between P sweeps
#define S_LM_MPPT_SWEEP     (S_LM_SAMPLES + 300)    //# run loops per SWEEP P sample
#endif


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Analog channel mapping:  channel=port=pin.  The values assigned to these
 * constants should match those used by ADC.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define ANLCAN0             0x00            //AN0=RA0=19
#define ANLCAN1             0x04            //AN1=RA1=18
#define ANLCAN2             0x08            //AN2=RA2=17
#define ANLCAN3             0x0C            //AN3=RA4=3
#define ANLCAN4             0x10            //AN4=RC0=16
#define ANLCAN5             0x14            //AN5=RC1=15
#define ANLCAN6             0x18            //AN6=RC2=14
#define ANLCAN7             0x1C            //AN7=RC3=7
#define ANLCAN8             0x20            //AN8=RC6=8
#define ANLCAN9             0x24            //AN9=RC7=9
#define ANLCAN10            0x28            //AN10=RB4=13
#define ANLCAN11            0x2C            //AN11=RB5=12
#define ANLCANS6            0x38            //Switched AN6
#define ANLCANS7            0x3C            //Switched AN7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PPS input pin definitions.  Map for MPU and use to assign peripheral input
 * pins.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPSRA0              0x00            //RA0
#define PPSRA1              0x01            //RA1
#define PPSRA2              0x02            //RA2
#define PPSRA3              0x03            //RA3
#define PPSRA4              0x04            //RA4
#define PPSRA5              0x05            //RA5
#define PPSRA6              0x06            //RA6
#define PPSRA7              0x07            //RA7
#define PPSRB0              0x08            //RB0
#define PPSRB1              0x09            //RB1
#define PPSRB2              0x0A            //RB2
#define PPSRB3              0x0B            //RB3
#define PPSRB4              0x0C            //RB4
#define PPSRB5              0x0D            //RB5
#define PPSRB6              0x0E            //RB6
#define PPSRB7              0x0F            //RB7
#define PPSRC0              0x10            //RC0
#define PPSRC1              0x11            //RC1
#define PPSRC2              0x12            //RC2
#define PPSRC3              0x13            //RC3
#define PPSRC4              0x14            //RC4
#define PPSRC5              0x15            //RC5
#define PPSRC6              0x16            //RC6
#define PPSRC7              0x17            //RC7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           PPS & direct pin mapping
 * 
 * For PPS output pin assignments both aliases for pin registers and constants 
 * mapping what peripheral connects to the pin.
 *      Pin register alias convention:  PPS_P_XXX_YY where
 *          XXX=peripheral name
 *          YY=channel or other detail; can be omitted
 *      Peripheral mapping convention:  PPS_M_XXX_YY same as alias
 * 
 * Input pins only require mapping constant.
 * 
 * For pins accessed directly or assigned in a peripheral's register (like ADC)
 * use the following convention:  PIN_XT_YYYYYY where
 *          X=(I)nput or (O)utput
 *          T=(A)nalog or (D)igital
 *          YYYYYY=user assigned mnemonic describing function
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Charger (COG) gate drive pin mapping
#define PPS_P_COG1_H    RB5PPS          //COG1 A maps to high side driver, pin 12
#define PPS_P_CLC1_L    RB6PPS          //CLC1 maps to low side B driver, pin 11
#define PPS_M_COG1_H    0x04            //Output COG1 A signal on selected pin
#define PPS_M_CLC1_L    0x01            //Output CLC1 on selected pin

//Output SSR (relay) control
#define PPS_P_PWM3_SSR  RB7PPS          //PWM3 maps to SSR gate drive on pin 10
#define PPS_M_PWM3_SSR  0x0E            //Output PWM3 signal on selected pin

//PRG2 rise / fall trigger pin mapping
#define PPS_M_PRG2_R    PPSRB5          //High side gate drive, pin 12
#define PPS_M_PRG2_F    PPSRB5          //High side gate drive, pin 12

#define PPS_M_LAT       0x00            //Remove peripheral from pin & output latch value

//Analog pin mapping
#define PIN_IA_PVAMP    ANLCAN8         //ADC sense - PV amps
#define PIN_IA_PVOLT    ANLCAN9         //ADC sense - PV volts
#define PIN_IA_OVOLT    ANLCAN4         //ADC sense - output volts
#define PIN_IA_BVOLT    ANLCAN2         //ADC sense - battery volts

//LED output pins for fault, run, and power levels
#define PIN_OD_FAULT    RC4             //Fault
#define PIN_OD_RUN      RC5             //Run


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Port/pin configuration:  I/O, analog/digital, slew rate, etc.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Port I/O:  1=input, 0=output
#define IO_PORTA    0b00000101
#define IO_PORTB    0b00010000
#define IO_PORTC    0b11001011
//Port Analog/Digital:  1=analog, 0=digital
#define AD_PORTA    0b00000100
#define AD_PORTB    0b00010000
#define AD_PORTC    0b11001111
//Port weak pull-up:  1=enabled, 0=disabled
#define WP_PORTA    0b00000000
#define WP_PORTB    0b00000000
#define WP_PORTC    0b00000000
//Open drain:  1=open drain, 0=push-pull
#define OD_PORTA    0b00000000
#define OD_PORTB    0b00000000
#define OD_PORTC    0b00000000
//Slew rate:  1=limited, 0=maximum
#define SR_PORTA    0b00000000
#define SR_PORTB    0b00000000
#define SR_PORTC    0b00000000
//Input level:  1=ST, 0=TLL
#define IL_PORTA    0b00000000
#define IL_PORTB    0b00000000
#define IL_PORTC    0b00000000
//Initial latch states:  all outputs off.
#define S_IN_PORTA  0b00000000
#define S_IN_PORTB  0b00000000
#define S_IN_PORTC  0b00000000


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * SYSTEM MACROS.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define setRelayOn()        PWM3DCH = CFG_PWM_3_DCH_ON
#define setRelayOff()       PWM3DCH = CFG_PWM_3_DCH_OFF
#define relayStatus()       ((PWM3DCH) ? TRUE : FALSE)
#define setledFAULT()       PIN_OD_FAULT = TRUE; PIN_OD_RUN = FALSE
#define setledRUN()         PIN_OD_FAULT = FALSE; PIN_OD_RUN = TRUE
#define cogStatus()         ((COG1ASD0bits.ASE) ? FALSE : TRUE)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Used by several console functions to remove whitespace between command and
 * parameter.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define rmvSpaces(cs)       for (cs += CMDSIZSTR; (*cs == S_CH_SPACE); ++cs)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * The getSnapshot macro is used by system events (NOT the console) to record a
 * snapshot of system status and set/respect the snapshot flag, flag_ss.
 * 
 * Only one snapshot can be taken and is not overwritten until flag_ss is set
 * FALSE via console command.  Only system initiated events set the flag.
 * Hence, a console command to take a snapshot will fail if the system has 
 * already saved one (and set flag_ss TRUE).  Likewise, if the console does
 * take a snapshot and a subsequent system event occurs before it is read, the
 * system snapshot will overwrite the console copy.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define getSnapshot()       if (!(flag_ss)) \
    memcpy((void *) &snapshot, (void *) &Config, sizeof(syscf)); flag_ss = TRUE


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FVR, DAC, and function Initialize configuration constants
 * DAC1:  output voltage set
 * DAC2:  current limit
 * DAC3:  Vdd (measured by ADC)
 * DAC4:  Synchronous low-side on-time Vref for CM3
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_FVR_C       0xCF            //Enabled, no temp sense, CDAFVRx4, \
                                          ADFVRx4
#define CFG_FVR_ADC_4X  0x03            //ADC buffer set to 4x gain
#define CFG_FVR_ADC_2X  0x02            //ADC buffer set to 2x gain


//DAC1:  voltage is set from NVM; default defined here
#define CFG_DAC_1_C     0x88            //Enabled, right justified, \
                                          FVR buffer 2, Vss
#define S_IN_VSET       865             //Loaded into DAC for output volts
#define S_IN_VSET_H     (S_IN_VSET >> S_SZ_BYTE)
#define S_IN_VSET_L     (S_IN_VSET & S_MK_U16_L)

//DAC2:  voltage is set from NVM; default defined here
#define CFG_DAC_2_C     0x88            //Enabled, right justified, \
                                          FVR buffer 2, Vss
#define S_IN_CLIM       500             //Loaded into DAC for current limiting
#define S_IN_CLIM_H     (S_IN_CLIM >> S_SZ_BYTE)
#define S_IN_CLIM_L     (S_IN_CLIM & S_MK_U16_L)

//DAC3: Vdd/4
#define CFG_DAC_3_C     0x80            //Enabled, (+)=Vdd, (-)=Vss
#define CFG_DAC_3_R     0x08            //Vdd / 4

//DAC4:  Vref timing threshold for synchronous low-side on-time
#define CFG_DAC_4_C     0x88            //Enabled, FVR buffer 2, Vss
#define CFG_DAC_4_R_LO  10              //1.28V ~ 640nS
#define CFG_DAC_4_R_HI  24              //3.072V ~ 1.536uS

#define FI_FVR_DLY      30              //Time delay in uS for FVR buffer to \
                                          stabilize


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *              ADC configuration, acquisition, and value storage
 * 
 * This section defines a uniform method for data acquisition and storage.  The
 * ADC is configured to operate thru function ReadVoltage and should not be 
 * accessed directly once function Initialize is called at boot.  Acquired 
 * values are stored in an object of type volts.  The integer and decimal 
 * portions of a voltage are stored in discrete variables.
 * 
 * Since the PIC family lacks floating point capability a 16-bit unsigned int
 * variable can represent the combined integer & decimal components in a format
 * referred to as 'VMV' (Volts-MilliVolts) format.  This allows evaluation as
 * an integer with the benefit of including the decimal portion for improved 
 * resolution.  This is done by multiplying the integer and dividing the decimal
 * values by some factors of 10, in effect shifting integer values left and
 * truncating decimal positions.
 * 
 * More places shifted offers higher resolution but with more limited range
 * (e.g. overflow).  The constants S_RT_VMV_INT & S_RT_VMV_DEC define the use of 
 * VMV values and must be assigned the paired values:
 * # Decimal Places         S_RT_VMV_INT        S_RT_VMV_DEC
 *          0               1                   1000
 *          1               10                  100
 *          2               100                 10
 *          3               1000                1
 * 
 * Prior to calling function ReadVoltage the variable mvr of the volts object
 * passed to the function must be initialized to either S_RT_ADC_MV_H or
 * S_RT_ADC_MV_L.  This defines the Vref (4.096 or 2.048) used and trades
 * resolution for voltage range.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//ADC peripheral configuration
#define CFG_ADC_CON_1       0xD3            //Result right justified, Fosc/16, Vss, FVR
#define CFG_ADC_CON_2       0x00            //No auto conversion

#define S_EN_ADC            0x01            //Enable bit position in ADCON0 register
#define S_LM_ADC_VAL        1023            //Maximum value 10-bit ADC can return
#define S_QY_MV_V           1000            //Number of mV in one volt
#define S_RT_ADC_MV_H       4               //mV / ADC bit for high range
#define S_RT_ADC_MV_L       2               //mV / ADC bit for low range
#define S_QY_ADC_1V_H       (S_QY_MV_V / S_RT_ADC_MV_H) //ADC value for 1V high
#define S_QY_ADC_1V_L       (S_QY_MV_V / S_RT_ADC_MV_L) //ADC value for 1V low
#define S_TM_ADC_ACQ        20              //# uS required for S&H to acquire
#define S_SZ_ADC_INT        sizeof(unsigned int)    //# bytes in ADC integer
#define POS_ADC_HB          1               //ADC result high byte
#define POS_ADC_LB          0               //ADC result low byte
#define S_RT_VMV_INT        10              //Sets integer position of VMV value
#define S_RT_VMV_DEC        100             //Sets decimal places of VMV value
#define S_RT_VMV_MUL        (S_RT_VMV_INT * 10) //Scales VMV * VMV result

//Internal channels (peripherals)
#define ADC_CH_FVR          0x7C            //FVR
#define ADC_CH_DAC_1        0x78            //DAC1
#define ADC_CH_TMP          0x74            //Temperature indicator
#define ADC_CH_DAC_2        0x70            //DAC2
#define ADC_CH_DAC_3        0x6C            //DAC3
#define ADC_CH_DAC_4        0x68            //DAC4


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * OPA1 configuration for outer voltage control loop
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OPA_1_C     0x80            //Enabled, no override or tristate
#define CFG_OPA_1_O     0x00            //Override not used
#define CFG_OPA_1_N     0x00            //(-) connect to OPA1IN0- pin 13
#define CFG_OPA_1_P     0x02            //(+) connect to DAC1_out


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 1 for current loop modulator (PWM) control
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_1_C0     0x8E            //Enabled, ZLF & hysteresis on
#define CFG_CM_1_C1     0x00            //No interrupts
#define CFG_CM_1_N      0x01            //(-) connects to IN1-
#define CFG_CM_1_P      0x08            //(+) connects to PRG1_output

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 2 for over-current sense
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_2_C0     0x8E            //Enabled, ZLF & hysteresis on
#define CFG_CM_2_C1     0x01            //Interrupt on negative edge of output
#define CFG_CM_2_N      0x01            //(-) connects to IN1-
#define CFG_CM_2_P      0x0B            //(+) connects to DAC2_output

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 3 for synchronous low-side timing
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_3_C0     0x8E            //Enabled, ZLF & hysteresis on
#define CFG_CM_3_C1     0x00            //No interrupts
#define CFG_CM_3_N      0x05            //(-) connects to PRG2 output
#define CFG_CM_3_P      0x0D            //(+) connects to DAC4_output


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Allocate timers to PWM modules
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CCP_TMRS    0x00            //PWM3, 4, CCP1 & 2 use timer2

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer2 configuration:  time base for PWM3.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_2_CLK   0x01            //Fosc/4
#define CFG_TMR_2_CON   0x80            //Enabled, 1:1 prescale, 1:1 post scale
#define CFG_TMR_2_HLT   0x00            //No sync, free running
#define CFG_TMR_2_RST   0x00            //No external reset sources
#define CFG_TMR_2_PR    39              //100kHz

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer6 configuration:  provides interrupt timing for background caretaker 
 * that performs system monitor & control.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_6_CLK   0x05            //MFINTOSC
#define CFG_TMR_6_CON   0x19            //Off, 1:2 prescale, 1:10 post scale
#define CFG_TMR_6_HLT   0xA0            //Prescale syncs Fosc/4, on bit syncs \
                                          to timer clock, free running
#define CFG_TMR_6_RST   0x00            //No external reset sources
#define CFG_TMR_6_PR    250             //Period value to obtain 10mS period


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure 10-bit PWM3:  gate drive for output SSR (relay)
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_3_CON           0x80            //Enabled, active high output
#define CFG_PWM_3_DCL           0x00            //2 LSB not used
#define CFG_PWM_3_DCH_OFF       0x00            //Relay off
#define CFG_PWM_3_DCH_ON        0x04            //Relay on, ~ 1.25uS DC

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure 16-bit PWM5:  master and sets max DC
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_5_CON           0x80            //Enabled, active high standard
#define CFG_PWM_5_INTE          0x00            //No interrupts
#define CFG_PWM_5_CLK           0x01            //No prescale, clock=HFINTOSC
#define CFG_PWM_5_LDCON         0x80            //Load triggering is immediate
#define CFG_PWM_5_OFCON         0x00            //Independent run mode
#define CFG_PWM_5_PHASE         0x0000          //Phase count
#define CFG_PWM_5_PERIOD        S_TM_MOD_F      //Period (frequency)
#define CFG_PWM_5_OFFSET        0x0000          //Offset count
#define CFG_PWM_5_TIMER         0x0000          //Starting value for timer


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure Complementary Output Generator 1 (COG1).
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_COG_1_CON0          0xD4            //Enable, load, HFINTOSC, HB
#define CFG_COG_1_CON1          0xC0            //Delay chain DB, active high

//Rising events
#define CFG_COG_1_RIS0          0x00            //None
#define CFG_COG_1_RIS1          0x02            //PWM5
#define CFG_COG_1_RSIM0         0x00            //None
#define CFG_COG_1_RSIM1         0x02            //PWM5

//Falling events
#define CFG_COG_1_FIS0          0x06            //CM1, CM2 - closed lopp
#define CFG_COG_1_FIS0_O        0x04            //CM2 - open lopp
#define CFG_COG_1_FIS1          0x02            //PWM5
#define CFG_COG_1_FSIM0         0x04            //Hi>low for CM2, low level CM1
#define CFG_COG_1_FSIM1         0x02            //Hi>low for PWM5

//Auto shutdown
#define CFG_COG_1_ASD0          0xA8            //Shutdown, outputs low
#define CFG_COG_1_ASD1          0x00            //No auto-shutdown sources

//Delays & blanking
#define CFG_COG_1_DBR           10              //Rising dead band
#define CFG_COG_1_DBF           15              //Falling dead band
#define CFG_COG_1_BLKR          0               //Rising edge blanking
#define CFG_COG_1_BLKF          0               //Falling edge blanking
#define CFG_COG_1_PHR           0               //Rising event phase delay
#define CFG_COG_1_PHF           0               //Falling event phase delay


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure CLC1 as AND gate for low-side synchronous drive
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CLC_1_CON           0x02            //Disabled, AND
#define CFG_CLC_1_POL           0x00            //No inverted polarities
#define CFG_CLC_1_SEL_0         0x0C            //COG1B for d1T/g1
#define CFG_CLC_1_SEL_1         0x0C            //COG1B for d2T/g2
#define CFG_CLC_1_SEL_2         0x09            //Sync C3 out
#define CFG_CLC_1_SEL_3         0x09            //Sync C3 out
#define CFG_CLC_1_GLS_0         0x02            //d1T is gated into g1
#define CFG_CLC_1_GLS_1         0x08            //d2T is gated into g2
#define CFG_CLC_1_GLS_2         0x20            //d3T is gated into g3
#define CFG_CLC_1_GLS_3         0x80            //d4T is gated into g4


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure PRG1 for slope compensation
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PRG_1_CON0          0xB0            //Enabled, edge, slope, off
#define CFG_PRG_1_CON1          0x02            //Fall low, rise high
#define CFG_PRG_1_INS           0x00            //OPA1OUT
#define CFG_PRG_1_CON2          0               //Slope=200mV/uS
#define CFG_PRG_1_RTSS          0x07            //PWM5
#define CFG_PRG_1_FTSS          0x07            //PWM5

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure PRG2 for low-side synchronous gate timing
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PRG_2_CON0          0xBA            //Enabled/edge/rise/one-shot/off
#define CFG_PRG_2_CON1          0x01            //Fall high, rise low
#define CFG_PRG_2_INS           0x08            //PRG2IN0 pin 7
#define CFG_PRG_2_CON2          0x1A            //Slope=2V/uS
#define CFG_PRG_2_RTSS          0x04            //PPS mapped to COG A out
#define CFG_PRG_2_FTSS          0x04            //PPS mapped to COG A out


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Print & Format Constants
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Object definitions
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Object to store a measured voltage with 3-digit mV resolution
typedef struct {
    unsigned int volt;                  //Voltage
    unsigned int mv;                    //Millivolts (3 decimals)
    unsigned int vmv;                   //Composite volts & mv represented as integer
    unsigned int adc;                   //Unmodified value read from ADC
    unsigned char mvr;                  //mV/ADC bit:  see ADC section for values
} volts;

//Configuration record to control system operation
//To calculate DC% from measured DC:  Config.odcp = ((Config.oldc * 100) / (CFG_PWM_5_PERIOD + 1));
typedef struct {
    uint8_t       sts;                  //System status code; see status codes
    uint8_t       ovcc;                 //Over-current counter
    uint8_t       pwr;                  //Calculated PV power
    uint8_t       mppt_pwr;             //Algorithm saved PV power
    uint8_t       mppt_dc;              //Algorithm assigned DC
    uint8_t       mppt_pdc;             //DC for peak recorded PV power
    unsigned int  pvco;                 //PV current offset from calibration
    volts         bat;                  //Battery (and Vcc supply) voltage
    volts         pvv;                  //PV voltage
    volts         pvc;                  //PV current
    volts         out;                  //Output voltage
    volts         vdd;                  //Logic (Vdd) supply voltage)
} syscf;


const char *CmdList[] = {
    {"CC"},                                 //Clear current counter
    {"CN"},                                 //Clear NVM error
    {"CP"},                                 //Clear snapshot
    {"GA"},                                 //Get PV amps
    {"GB"},                                 //Get battery volts
    {"GC"},                                 //Get over-current count
    {"GD"},                                 //Get set duty cycle %
    {"GE"},                                 //Get set output volts
    {"GF"},                                 //Get charger ofF @ PV< voltage
    {"GG"},                                 //Get charger Go @ PV> voltage
    {"GI"},                                 //Get set over-current limit
    {"GL"},                                 //Get loop setting
    {"GM"},                                 //Get MPPT operating duty cycle %
    {"GN"},                                 //Get NVM status code
    {"GO"},                                 //Get output volts
    {"GP"},                                 //Get snapshot of dynamic parameters
    {"GR"},                                 //Get program release
    {"GS"},                                 //Get system status
    {"GT"},                                 //Get autostart setting
    {"GV"},                                 //Get PV volts
    {"GW"},                                 //Get PV Watts
    {"GY"},                                 //Get relay status
    {S_CD_GETD},                            //Get data payload, see dspstatus.h
    {"RD"},                                 //Restore NVM defaults
    {"RT"},                                 //Reset MCU
    {"SA"},                                 //Set autostart 1 | 0
    {"SC"},                                 //Set current limit #
    {"SD"},                                 //Set max / open-loop duty cycle %
    {"SE"},                                 //Set echo 1 | 0
    {"SF"},                                 //Set charger ofF @ PV< voltage
    {"SG"},                                 //Set charger Go @ PV> voltage
    {"SL"},                                 //Set loop 1 | 0
    {"SO"},                                 //Set charger on
    {"SV"},                                 //Set output volts #
    {"SY"}                                  //Set relay on
};
#define NUMOFCMDS   sizeof(CmdList) / sizeof(const char *)
#define CMDCLRCNT   0                       //Clear current counter
#define CMDCLRNVM   1                       //Clear NVM error
#define CMDCLRSNP   2                       //Clear snapshot
#define CMDGETAMP   3                       //Get PV amps
#define CMDGETBAT   4                       //Get battery volts
#define CMDGETCNT   5                       //Get over-current count
#define CMDGETDCY   6                       //Get set duty cycle %
#define CMDGETOVS   7                       //Get set output volts
#define CMDGETPVL   8                       //Get charger ofF @ PV< voltage
#define CMDGETPVG   9                       //Get charger Go @ PV> voltage
#define CMDGETCUR   10                      //Get set over-current limit
#define CMDGETLPS   11                      //Get loop setting
#define CMDGETMDC   12                      //Get MPPT operating duty cycle %
#define CMDGETNVM   13                      //Get NVM status code
#define CMDGETOUT   14                      //Get output volts
#define CMDGETSNP   15                      //Get snapshot of dynamic parameters
#define CMDGETREL   16                      //Get program release
#define CMDGETSTS   17                      //Get system status
#define CMDGETAUT   18                      //Get autostart setting
#define CMDGETVOL   19                      //Get PV volts
#define CMDGETWAT   20                      //Get PV Watts
#define CMDGETRLY   21                      //Get relay status
#define CMDGETDAT   22                      //Get data payload for rs232-44780
#define CMDRSTDFL   23                      //Restore NVM defaults
#define CMDRSTMCU   24                      //Reset MCU
#define CMDSETAUT   25                      //Set autostart 1 | 0
#define CMDSETCUR   26                      //Set current limit #
#define CMDSETDCY   27                      //Set max / open-loop duty cycle %
#define CMDSETECH   28                      //Set echo 1 | 0
#define CMDSETPVL   29                      //Set charger ofF @ PV< voltage
#define CMDSETPVG   30                      //Set charger Go @ PV> voltage
#define CMDSETLOP   31                      //Set loop 1 | 0
#define CMDSETCON   32                      //Set charger on
#define CMDSETVOL   33                      //Set output volts #
#define CMDSETRLY   34                      //Set relay on

#define CMDSIZSTR   2                       //Size of command string


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Message, format, and label strings
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char *msgAck = "Ack\n",               //Command completed
           *msgNck = "Nck\n",               //Command completed but unsuccessful
           *fmtGpib = "%u.%u\n",            //GPIB inquiry data response format
           *fmtPrg = "%u.%u b%u\n";         //Firmware version response format


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Flash program memory definitions.  For non-volatile storage of configurable
 * parameters.
 * 
 * Word values must be stored in little endian (L|H) format.
 * 
 * Important:  flash architecture is PIC family dependent and the memory map /
 * constants must be reviewed when changing MCUs.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define FLASHROWSZ          32              //Size of flash row: PIC specific
#define FLASH_SZ_CSUM   (FLASHROWSZ - 2)    //Store checksum in last word of row

//Flash memory unlock sequence macro
#define unlockSeq() PMCON2 = 0x55; PMCON2 = 0xAA; WR = TRUE; NOP(); NOP()

//Row address for data:  this value must match default assignment below
#define FLASH_AR_AUTO       0               //Autostart if TRUE
#define FLASH_AR_STS        1               //Last status code
#define FLASH_AR_DC         2               //Max (PWM5) or open loop duty cycle
#define FLASH_AR_LOOP       3               //True if closed
#define FLASH_AR_CLIM_L     4               //Current limit (DAC2) low byte
#define FLASH_AR_CLIM_H     5               //Current limit (DAC2) high byte
#define FLASH_AR_CLIM       (FLASH_AR_CLIM_L / sizeof(uint16_t))
#define FLASH_AR_VSET_L     6               //Output volts (DAC1) low byte
#define FLASH_AR_VSET_H     7               //Output volts (DAC1) high byte
#define FLASH_AR_VSET       (FLASH_AR_VSET_L / sizeof(uint16_t))
#define FLASH_AR_STATUS     8               //STATUS register
#define FLASH_AR_RDB        9               //Rising dead-band count
#define FLASH_AR_FDB        10              //Falling dead-band count
#define FLASH_AR_RAMP       11              //Slope comp setting
#define FLASH_AR_PCON       12              //PCON register
#define FLASH_AR_PVVL_STOP  13              //PV volts to stop charger
#define FLASH_AR_PVVL_RSRT  14              //PV volts to restart charger

//Checksum stored in little endian format in last word of row
#define FLASH_AR_CSUM_L (FLASHROWSZ - 2)
#define FLASH_AR_CSUM_H (FLASHROWSZ - 1)
#define FLASH_AR_CSUM   (FLASH_AR_CSUM_L / sizeof(uint16_t))

//Default data to initialize row with.  Positions 30 & 31 reserved for checksum.
#define S_DF_FLASH_0        TRUE
#define S_DF_FLASH_1        STS_UND
#define S_DF_FLASH_2        S_DC_RUN
#define S_DF_FLASH_3        TRUE
#define S_DF_FLASH_4        S_IN_CLIM_L
#define S_DF_FLASH_5        S_IN_CLIM_H
#define S_DF_FLASH_6        S_IN_VSET_L
#define S_DF_FLASH_7        S_IN_VSET_H
#define S_DF_FLASH_8        0
#define S_DF_FLASH_9        CFG_COG_1_DBR
#define S_DF_FLASH_10       CFG_COG_1_DBF
#define S_DF_FLASH_11       CFG_PRG_1_CON2
#define S_DF_FLASH_12       0
#define S_DF_FLASH_13       S_VL_PV_STOP
#define S_DF_FLASH_14       S_VL_PV_RSRT
#define S_DF_FLASH_15       0
#define S_DF_FLASH_16       0
#define S_DF_FLASH_17       0
#define S_DF_FLASH_18       0
#define S_DF_FLASH_19       0
#define S_DF_FLASH_20       0
#define S_DF_FLASH_21       0
#define S_DF_FLASH_22       0
#define S_DF_FLASH_23       0
#define S_DF_FLASH_24       0
#define S_DF_FLASH_25       0
#define S_DF_FLASH_26       0
#define S_DF_FLASH_27       0
#define S_DF_FLASH_28       0
#define S_DF_FLASH_29       0


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations.  See source file for explanation of each function.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void main(void);
uint8_t ChargerON(void);
void ChargerOFF(void);
void ReadVoltage(unsigned char, unsigned char, volts *);
void CalcAvg(unsigned int, unsigned char, unsigned char, volts *);
void chgDutyCycle(uint16_t, uint16_t);
void eraseFlashRow(uint16_t addr);
void writeFlashRow(uint16_t, unsigned char *);
uint16_t nvmChksum(unsigned char *);
void interrupt InterruptExec(void);
void Initialize(void);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Preprocessor checks.  Maintain at end of main include file.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Check for sum of samples exceeding unsigned int
#define _ADC_MAX_VAL    S_IN_CLIM
#if(_ADC_MAX_VAL < (((S_LM_PVV / S_RT_PVV) * 1000) / S_RT_ADC_MV_H))
#define _ADC_MAX_VAL    (((S_LM_PVV / S_RT_PVV) * 1000) / S_RT_ADC_MV_H)
#endif
#if((_ADC_MAX_VAL * S_LM_SAMPLES) > UINT_MAX)
#error "Sum of samples exceeds unsigned int."
#endif



//End of file